/* recCostar.c - record support routines for the COSTAR chip (STAR/SSD) */
/*
 *	Institut de Recherche Subatomique de Strasbourg
 *
 *	Author:		David Bonnet
 *	Date:		22/03/00
 */ 

#include	<vxWorks.h>
#include 	<taskLib.h>
#include 	<usrLib.h>
#include	<types.h>
#include	<stdioLib.h>
#include	<lstLib.h>
#include	<string.h>

#include	<alarm.h>
#include	<dbDefs.h>
#include	<dbAccess.h>
#include	<dbScan.h>
#include	<dbFldTypes.h>
#include	<devSup.h>
#include	<errMdef.h>
#include	<recSup.h>
#include 	<jtagCommon.h>
#include 	<jtagCostar.h>
#define GEN_SIZE_OFFSET
#include 	<costarRecord.h>
#undef GEN_SIZE_OFFSET

/* Create RSET - Record Support Entry Table*/
#define report NULL
static long initialize();
static long init_record();
static long process();
#define special NULL
static long get_value(); 
#define cvt_dbaddr NULL
#define get_array_info NULL
#define put_array_info NULL
static long get_units();
static long get_precision();
static long get_enum_str(); 
static long get_enum_strs(); 
static long put_enum_str(); 
static long get_graphic_double();
static long get_control_double();
#define get_alarm_double NULL


struct rset costarRSET={
	RSETNUMBER,
	report,
	initialize,
	init_record,
	process,
	special,
	get_value,
	cvt_dbaddr,
	get_array_info,
	put_array_info,
	get_units,
	get_precision,
	get_enum_str,
	get_enum_strs,
	put_enum_str,
	get_graphic_double,
	get_control_double,
	get_alarm_double};

int async_costar(void);
int restart_async_costar();
static long costar_io();
static void monitor();
static void alarm();

static char INIT_JTAG[LADDER_NB];
static unsigned short HB_MAP[LADDER_NB];
costar 	costar_list[LADDER_NB][MODULE_NB];
unsigned char costar_list_old_vref[LADDER_NB][MODULE_NB];
unsigned short module_mux[LADDER_NB][MODULE_NB]; 

static long initialize()
{	
	int i;
	for (i=0;i<LADDER_NB;i++) INIT_JTAG[i]=0;
	return 0;
}

static long init_record(costarRecord *pcostar,int pass)
{
	float vrn,vrp;
	
 	if (pass==0)
	{
	vrn=pcostar->vrn;
	vrp=pcostar->vrp;
	costar_list[pcostar->ladr][pcostar->modu].vref=256*(pcostar->refv+2.0-vrn)/(vrp-vrn);
	costar_list_old_vref[pcostar->ladr][pcostar->modu]=127;
	costar_list[pcostar->ladr][pcostar->modu].soft_bypass=pcostar->byps;
	costar_list[pcostar->ladr][pcostar->modu].hard_bypass=pcostar->hbyp;
	pcostar->jini=3;
	}

	return 0;
}

static long process(costarRecord *pcostar)
{
	long status;
	
	status=costar_io(pcostar); /* Do costar I/O operation */
	
	pcostar->pact=TRUE;

	alarm(pcostar);
	monitor(pcostar);
	recGblGetTimeStamp(pcostar);
	recGblFwdLink(pcostar);
	pcostar->pact=FALSE;
	return(status);
}	

static long get_value(costarRecord *pcostar,struct valueDes *pvdes)
{
    pvdes->field_type = DBF_ENUM;
    pvdes->no_elements=1;
    (unsigned short *)(pvdes->pvalue) = &pcostar->val;
    return(0);
}

static long get_units(struct dbAddr *paddr,char *units)
{
    	costarRecord	*pcostar=(costarRecord *)paddr->precord;
     	
	if(paddr->pfield==(void *)&pcostar->avdd || 
           paddr->pfield==(void *)&pcostar->avss ||
	   paddr->pfield==(void *)&pcostar->refv || 
           paddr->pfield==(void *)&pcostar->vrp || 
           paddr->pfield==(void *)&pcostar->vrn)
    	
		strncpy(units,"V",2); 
	
	else if (paddr->pfield==(void *)&pcostar->bias || 
                 paddr->pfield==(void *)&pcostar->guar)
		 	
		strncpy(units,"uA",3);
			
	else if (paddr->pfield==(void *)&pcostar->val || 
	          paddr->pfield==(void *)&pcostar->cfb)
	
		strncpy(units,"C",3); 
	
	else if (paddr->pfield==(void *)&pcostar->cfa)
	
		strncpy(units,"C/adc",7); 
	
	else if (paddr->pfield==(void *)&pcostar->ires)
	
		strncpy(units,"Kohm",5); 
	
        else units[0]=0;
	 
    return(0);
}

static long get_precision(paddr,precision)
    struct dbAddr *paddr;
    long	  *precision;
{
    	costarRecord	*pcostar=(costarRecord *)paddr->precord;
	
     	if((paddr->pfield==(void *)&pcostar->avdd) || 
           (paddr->pfield==(void *)&pcostar->avss) ||
	   (paddr->pfield==(void *)&pcostar->val))
		*precision=1;
	else if ((paddr->pfield==(void *)&pcostar->bias) || 
                 (paddr->pfield==(void *)&pcostar->guar))
		*precision=2;
	else if ((paddr->pfield==(void *)&pcostar->refv) || 
                 (paddr->pfield==(void *)&pcostar->vrn) ||
	         (paddr->pfield==(void *)&pcostar->vrp) ||
		 (paddr->pfield==(void *)&pcostar->cfa) || 
                 (paddr->pfield==(void *)&pcostar->cfb))
		*precision=3; 
    	else recGblGetPrec(paddr,precision);
    return(0);
}

static long get_graphic_double(struct dbAddr *paddr,struct dbr_grDouble	*pgd)
{
    	costarRecord	*pcostar=(costarRecord *)paddr->precord;
	
     	if(paddr->pfield==(void *)&pcostar->avdd)
	{ 
         	pgd->upper_disp_limit = 3.0; 
         	pgd->lower_disp_limit = 0.0;
	} 
     	if(paddr->pfield==(void *)&pcostar->avss)
	{ 
         	pgd->upper_disp_limit = 0.0; 
         	pgd->lower_disp_limit = -3.0;
	} 
	else if (paddr->pfield==(void *)&pcostar->bias)
	{
         	pgd->upper_disp_limit = 7; 
         	pgd->lower_disp_limit = 0;
	}
	else if (paddr->pfield==(void *)&pcostar->guar)
	{
         	pgd->upper_disp_limit = 5; 
         	pgd->lower_disp_limit = 0;
	} 
	else if (paddr->pfield==(void *)&pcostar->val)
	{
          	pgd->upper_disp_limit = 80.0; 
          	pgd->lower_disp_limit = -20.0;
	}
        else recGblGetGraphicDouble(paddr,pgd);
	 
    	return(0);
}

static long get_control_double(struct dbAddr *paddr,struct dbr_ctrlDouble *pcd)
{
    	costarRecord	*pcostar=(costarRecord *)paddr->precord;
	
	
     	if(paddr->pfield==(void *)&pcostar->refv)
	{ 
         	pcd->upper_ctrl_limit = 1.0; 
         	pcd->lower_ctrl_limit = -1.0;
	} 
	else if (paddr->pfield==(void *)&pcostar->vrn || 
                 paddr->pfield==(void *)&pcostar->vrp)
	{
         	pcd->upper_ctrl_limit = 4.0;
         	pcd->lower_ctrl_limit = 0.0;
	} 

        else recGblGetControlDouble(paddr,pcd);
	 
    	return(0);
}

static long get_enum_str(paddr,pstring)
    struct dbAddr *paddr;
    char	  *pstring;
{
    struct costarRecord	*pcostar=(struct costarRecord *)paddr->precord;

    if(paddr->pfield==(void *)&pcostar->byps)
    {
       if(pcostar->byps==0) {
	  strcpy(pstring,"OFF");
       } else if(pcostar->byps==1) {
	  strcpy(pstring,"ON");
       } else {
	  strcpy(pstring,"Illegal_value");
       }
    } else if(paddr->pfield==(void *)&pcostar->hbyp)
    {
       if(pcostar->hbyp==0) {
	  strcpy(pstring,"OFF");
       } else if(pcostar->hbyp==1) {
	  strcpy(pstring,"ON");
       } else {
	  strcpy(pstring,"Illegal_value");
       }
    } 
    return(0);
}

static long get_enum_strs(paddr,pes)
    struct dbAddr *paddr;
    struct dbr_enumStrs *pes;
{
    struct costarRecord	*pcostar=(struct costarRecord *)paddr->precord;

    pes->no_str = 2;
    memset(pes->strs,'\0',sizeof(pes->strs));
    strcpy(pes->strs[0],"OFF");
    strcpy(pes->strs[1],"ON");   
    return(0);   
}

static long put_enum_str(paddr,pstring)
    struct dbAddr *paddr;
    char          *pstring;
{
    struct costarRecord     *pcostar=(struct costarRecord *)paddr->precord;

    if(paddr->pfield==(void *)&pcostar->byps)
    {
       if(strcmp(pstring,"OFF")==0) pcostar->byps = 0;
       else  if(strcmp(pstring,"ON")==0) pcostar->byps = 1;
       else return(S_db_badChoice);
    } else if(paddr->pfield==(void *)&pcostar->hbyp)
    {
       if(strcmp(pstring,"OFF")==0) pcostar->hbyp = 0;
       else  if(strcmp(pstring,"ON")==0) pcostar->hbyp = 1;
       else return(S_db_badChoice);
    }
    
    return(0);
}

static void alarm(costarRecord *pcostar)
{
	float		vmon,vhh,vhi,vll,vlo,vhys,vlal;
	float		bias,bhh,bhi,bll,blo,bhys,blal;
	float		guar,ghh,ghi,gll,glo,ghys,glal;
	float		temp,thh,thi,thys,tlal;
	unsigned short	vhhs,vhs,vlls,vls;
	unsigned short	bhhs,bhs,blls,bls;
	unsigned short	ghhs,ghs,glls,gls;
	unsigned short	thhs,ths;
	unsigned char 	alarm_status;
	
	vhh = pcostar->vhh; vll = pcostar->vll; vhi = pcostar->vhi; vlo = pcostar->vlo;
	bhh = pcostar->bhh; bll = pcostar->bll; bhi = pcostar->bhi; blo = pcostar->blo;
	ghh = pcostar->ghh; gll = pcostar->gll; ghi = pcostar->ghi; glo = pcostar->glo;
	thh = pcostar->thh; thi = pcostar->thi;
	vhhs = pcostar->vhhs; vlls = pcostar->vlls; vhs = pcostar->vhs; vls = pcostar->vls;
	bhhs = pcostar->bhhs; blls = pcostar->blls; bhs = pcostar->bhs; bls = pcostar->bls;
	ghhs = pcostar->ghhs; glls = pcostar->glls; ghs = pcostar->ghs; gls = pcostar->gls;
	thhs = pcostar->thhs;ths = pcostar->ths;
	vmon = pcostar->avdd-pcostar->avss; vhys = pcostar->vhys; vlal = pcostar->vlal;
	bias = pcostar->bias; bhys = pcostar->bhys; blal = pcostar->blal;
	guar = pcostar->guar; ghys = pcostar->ghys; glal = pcostar->glal;
	temp = pcostar->val; thys = pcostar->thys; tlal = pcostar->tlal;
	
	pcostar->alst=0;
	
	if ((pcostar->jini==0)||(pcostar->byps==0)||(pcostar->hbyp==0)) 
	{
	   /*printf("i:%d b:%d h:%d\n",INIT_JTAG[pcostar->ladr],pcostar->byps,pcostar->hbyp);*/
	   
	   if((pcostar->byps==0)||(pcostar->hbyp==0)) pcostar->alst=15;
	   /*if(pcostar->jini==0) recGblSetSevr(pcostar,STATE_ALARM,INVALID_ALARM);*/
	   pcostar->avss=0.0;
	   pcostar->avdd=0.0;
	   pcostar->val=0.0;
	   pcostar->bias=0.0;
	   pcostar->guar=0.0;
	   return;
	}
	
	/* alarm condition hihi */
/*	if (vhhs && (vmon >= vhh || ((vlal==vhh) && (vmon >= vhh-vhys)))){
	        if (recGblSetSevr(pcostar,HIHI_ALARM,pcostar->vhhs)) pcostar->vlal = vhh;
		pcostar->alst|=1;
	}
	if (bhhs && (bias >= bhh || ((blal==bhh) && (bias >= bhh-bhys)))){
	        if (recGblSetSevr(pcostar,HIHI_ALARM,pcostar->bhhs)) pcostar->blal = bhh;
		pcostar->alst|=2;
	}
	if (ghhs && (guar >= ghh || ((glal==ghh) && (guar >= ghh-ghys)))){
	        if (recGblSetSevr(pcostar,HIHI_ALARM,pcostar->ghhs)) pcostar->glal = ghh;
		pcostar->alst|=4;
	}
	if (thhs && (temp >= thh || ((tlal==thh) && (temp >= thh-thys)))){
	        if (recGblSetSevr(pcostar,HIHI_ALARM,pcostar->thhs)) pcostar->tlal = thh;
		pcostar->alst|=8;
	}*/

	/* alarm condition lolo */
	/*if (!(pcostar->alst&1) && vlls && (vmon <= vll || ((vlal==vll) && (vmon <= vll+vhys)))){
	        if (recGblSetSevr(pcostar,LOLO_ALARM,pcostar->vlls)) pcostar->vlal = vll;
		pcostar->alst|=1;
	}
	if (!(pcostar->alst&2) && blls && (bias <= bll || ((blal==bll) && (bias <= bll+bhys)))){
	        if (recGblSetSevr(pcostar,LOLO_ALARM,pcostar->blls)) pcostar->blal = bll;
		pcostar->alst|=2;
	}
	if (!(pcostar->alst&4) && glls && (guar <= gll || ((glal==gll) && (guar <= gll+ghys)))){
	        if (recGblSetSevr(pcostar,LOLO_ALARM,pcostar->glls)) pcostar->glal = gll;
		pcostar->alst|=4;
	}*/

	/* alarm condition high */
	/*if (!(pcostar->alst&1) && vhs && (vmon >= vhi || ((vlal==vhi) && (vmon >= vhi-vhys)))){
	        if (recGblSetSevr(pcostar,HIGH_ALARM,pcostar->vhs)) pcostar->vlal = vhi;
		pcostar->alst|=1;
	}
	if (!(pcostar->alst&2) && bhs && (bias >= bhi || ((blal==bhi) && (bias >= bhi-bhys)))){
	        if (recGblSetSevr(pcostar,HIGH_ALARM,pcostar->bhs)) pcostar->blal = bhi;
		pcostar->alst|=2;
	}
	if (!(pcostar->alst&4) && ghs && (guar >= ghi || ((glal==ghi) && (guar >= ghi-ghys)))){
	        if (recGblSetSevr(pcostar,HIGH_ALARM,pcostar->ghs)) pcostar->glal = ghi;
		pcostar->alst|=4;
	}
	if (!(pcostar->alst&8) && ths && (temp >= thi || ((tlal==thi) && (temp >= thi-thys)))){
	        if (recGblSetSevr(pcostar,HIGH_ALARM,pcostar->ths)) pcostar->tlal = thi;
		pcostar->alst|=8;
	}*/

	/* alarm condition low */
	/*if (!(pcostar->alst&1) && vls && (vmon <= vlo || ((vlal==vlo) && (vmon <= vlo+vhys)))){
	        if (recGblSetSevr(pcostar,LOW_ALARM,pcostar->vls)) pcostar->vlal = vlo;
		pcostar->alst|=1;
	}
	if (!(pcostar->alst&2) && bls && (bias <= blo || ((blal==blo) && (bias <= blo+bhys)))){
	        if (recGblSetSevr(pcostar,LOW_ALARM,pcostar->bls)) pcostar->blal = blo;
		pcostar->alst|=2;
	}  
	if (!(pcostar->alst&4) && gls && (guar <= glo || ((glal==glo) && (guar <= glo+ghys)))){
	        if (recGblSetSevr(pcostar,LOW_ALARM,pcostar->gls)) pcostar->glal = glo;
		pcostar->alst|=4;
	} */ 
	
		
	
	/* we get here only if val is out of alarm by at least hyst */
	if(pcostar->alst==0)
	{
	   pcostar->vlal = vmon;
	   pcostar->blal = bias;
	   pcostar->glal = guar;
	   pcostar->tlal = temp;
	}
	return;
}

static void monitor(costarRecord *pcostar)
{
	unsigned short	monitor_mask,mm;
	float		delta;

	monitor_mask = mm = recGblResetAlarms(pcostar);
        /* check for alarm status change */
	if(pcostar->alst!=pcostar->lals){
		monitor_mask |= DBE_VALUE | DBE_LOG;
		pcostar->lals=pcostar->alst;
	}
	/* send out monitors connected to the value field */
	if (monitor_mask)
		db_post_events(pcostar,&pcostar->alst,monitor_mask);
		
        if((!pcostar->byps)||(!pcostar->hbyp)) return;
	
	monitor_mask = mm;
	/* check for temperature value change */
	delta = pcostar->ltp - pcostar->val;
	if(delta<0.0) delta = -delta;
	if (delta > pcostar->tmde) {
		/* post events for value change */
		monitor_mask |= DBE_VALUE;
		/* update last value monitored */
		pcostar->ltp = pcostar->val;
	}

	/* check for temperature archive value change */
	delta = pcostar->tlar - pcostar->val;
	if(delta<0.0) delta = -delta;
	if (delta > pcostar->tade) {
		/* post events for value change */
		monitor_mask |= DBE_LOG;
		/* update last value monitored */
		pcostar->tlar = pcostar->val;
	}

	/* send out monitors connected to the value field */
	if (monitor_mask)
		db_post_events(pcostar,&pcostar->val,monitor_mask);
	
        monitor_mask = mm;
	/* check for avdd value change */
	delta = pcostar->lvdd - pcostar->avdd;
	if(delta<0.0) delta = -delta;
	if (delta > pcostar->vmde) {
		/* post events for value change */
		monitor_mask |= DBE_VALUE;
		/* update last value monitored */
		pcostar->lvdd = pcostar->avdd;
	}

	/* check for avdd archive value change */
	delta = pcostar->dlar - pcostar->avdd;
	if(delta<0.0) delta = -delta;
	if (delta > pcostar->vade) {
		/* post events for value change */
		monitor_mask |= DBE_LOG;
		/* update last value monitored */
		pcostar->dlar = pcostar->avdd;
	}

	/* send out monitors connected to the value field */
	if (monitor_mask)
		db_post_events(pcostar,&pcostar->avdd,monitor_mask);
		
        monitor_mask = mm;
	/* check for avss value change */
	delta = pcostar->lvss - pcostar->avss;
	if(delta<0.0) delta = -delta;
	if (delta > pcostar->vmde) {
		/* post events for value change */
		monitor_mask |= DBE_VALUE;
		/* update last value monitored */
		pcostar->lvss = pcostar->avss;
	}

	/* check for avss archive value change */
	delta = pcostar->slar - pcostar->avss;
	if(delta<0.0) delta = -delta;
	if (delta > pcostar->vade) {
		/* post events for value change */
		monitor_mask |= DBE_LOG;
		/* update last value monitored */
		pcostar->slar = pcostar->avss;
	}

	/* send out monitors connected to the value field */
	if (monitor_mask)
		db_post_events(pcostar,&pcostar->avss,monitor_mask);
        
	monitor_mask = mm;
	/* check for bias current value change */
	delta = pcostar->lbia - pcostar->bias;
	if(delta<0.0) delta = -delta;
	if (delta > pcostar->bmde) {
		/* post events for value change */
		monitor_mask |= DBE_VALUE;
		/* update last value monitored */
		pcostar->lbia = pcostar->bias;
	}

	/* check for bias current archive value change */
	delta = pcostar->blar - pcostar->bias;
	if(delta<0.0) delta = -delta;
	if (delta > pcostar->bade) {
		/* post events for value change */
		monitor_mask |= DBE_LOG;
		/* update last value monitored */
		pcostar->blar = pcostar->bias;
	}

	/* send out monitors connected to the value field */
	if (monitor_mask)
		db_post_events(pcostar,&pcostar->bias,monitor_mask);
        
	monitor_mask = mm;
	/* check for guard current value change */
	delta = pcostar->lgua - pcostar->guar;
	if(delta<0.0) delta = -delta;
	if (delta > pcostar->gmde) {
		/* post events for value change */
		monitor_mask |= DBE_VALUE;
		/* update last value monitored */
		pcostar->lgua = pcostar->guar;
	}

	/* check for guard current value change */
	delta = pcostar->glar - pcostar->guar;
	if(delta<0.0) delta = -delta;
	if (delta > pcostar->gade) {
		/* post events for value change */
		monitor_mask |= DBE_LOG;
		/* update last value monitored */
		pcostar->glar = pcostar->guar;
	}

	/* send out monitors connected to the value field */
	if (monitor_mask)
		db_post_events(pcostar,&pcostar->guar,monitor_mask);
	
	
	return;
}


static long costar_io(costarRecord *pcostar)
{
	unsigned char temp,bias,guard,avdd,avss;
	float vrn,vrp,CoefA,CoefB,fvss;

		temp=costar_list[pcostar->ladr][pcostar->modu].adc1[3];
		avdd=costar_list[pcostar->ladr][pcostar->modu].adc1[2];
		avss=costar_list[pcostar->ladr][pcostar->modu].adc1[1];
		bias=costar_list[pcostar->ladr][pcostar->modu].adc0[2];
		guard=costar_list[pcostar->ladr][pcostar->modu].adc0[1];
		/*printf("t:%d +:%d -:%d b:%d
		g:%d\n",temp,avdd,avss,bias,guard);*/
		costar_list[pcostar->ladr][pcostar->modu].soft_bypass=pcostar->byps;
		pcostar->hbyp=costar_list[pcostar->ladr][pcostar->modu].hard_bypass;
		module_mux[pcostar->ladr][pcostar->modu]=(pcostar->muxb&3)
							+((pcostar->muxa&3)<<2)
							+((pcostar->muxp&7)<<4)
							+((pcostar->muxm&7)<<7);
		vrn=pcostar->vrn;
		vrp=pcostar->vrp;
		CoefA=pcostar->cfa;
		CoefB=pcostar->cfb;
		pcostar->val=temp*CoefA+CoefB;
		pcostar->avss=fvss=-(avss*(vrp-vrn)/256+vrn)/0.6;
		pcostar->avdd=(avdd*(vrp-vrn)/256+vrn)/0.4+fvss;
		pcostar->bias=(bias*(vrp-vrn)/256+vrn+fvss)*1000/pcostar->ires;
		pcostar->guar=(guard*(vrp-vrn)/256+vrn+fvss)*1000/pcostar->ires;
		if(pcostar->guar<0.0) pcostar->guar=-pcostar->guar;
		if(pcostar->bias<0.0) pcostar->bias=-pcostar->bias;
		pcostar->leak=pcostar->bias+pcostar->guar;
		costar_list[pcostar->ladr][pcostar->modu].vref=256*(pcostar->refv+2.0-vrn)/(vrp-vrn);
		if (pcostar->jini!=3)
		   INIT_JTAG[pcostar->ladr]=pcostar->jini;
		return 1;
}
