/* jtagCommon.c		Jtag common functions 
 *
 *	Institut de Recherche Subatomique de Strasbourg
 *
 *	Author :	David Bonnet
 *	Date :		22/03/00
 *
 */

/*  

8990 function prototypes:

void write_tbc(unsigned short register_offset,unsigned short data)
void read_tbc(unsigned short register_offset,unsigned short *data)
void write_buf(unsigned short data)
void read_buf(unsigned short *data)
void wait_jtag_state(short state)
unsigned short jtag_state(void)
short hard_reset(unsigned short target,unsigned short clock_source,unsigned short divider)
void soft_reset()
void scan_ir(unsigned short *output,unsigned short length,unsigned short *input)
void scan_dr(unsigned short *output,unsigned short length,unsigned short *input)
void circulate_dr(unsigned short length,unsigned short *output)
short reset_jtag()
short reset_jtag_var(unsigned short i,unsigned short j,unsigned short k)
void connect_tap(char tap)
void disconnect_tap(char tap)
void bypass_all_chips(chip *map)
int jtag_check(unsigned short ir)
unsigned short detect_length()
void clear_bypass()
int auto_init_map(chip *map)
int jtag_init(chip *map,char *chain)
*/


#include <string.h> 
#include "vxWorks.h"
#include "sysLib.h"
#include "stdio.h"
#include "jtagCommon.h"


int JTAG_INIT_DONE=0;
unsigned short logical_address = 32;
unsigned short a16base=0xc800;
unsigned char tck_delay = 0;
chip tcnx[3]={CONNEXION,READOUT,END};
chip tread[2]={READOUT,END};
chip thyb[10]={COSTAR,ALICE128C,ALICE128C,ALICE128C,ALICE128C,ALICE128C,ALICE128C,CONNEXION,READOUT,END};

/*************************************************************************

       Function: writea16()

        Summary: Writes a 16-bit value to VMEbus A16 space

          Usage: short writea16(addr, data)
                 unsigned short addr;
                 unsigned short data;

    Description: This routine writes the specified 16-bit value to the
                 register offset specified by addr.

*************************************************************************/

short writea16(unsigned short addr,unsigned short data)
{
    unsigned short *ptr = NULL;

    /* create pointer to A16 space of CVME-1149.1 */
    
   sysBusToLocalAdrs(0x29,(char*)addr+a16base,(char **) &ptr);
   
    *ptr = data;            /* write data */

    return 0;
}

/*************************************************************************

       Function: reada16()

        Summary: Reads a 16-bit value from VMEbus A16 space

          Usage: short writea16(addr, readval)
                 unsigned short addr;
                 unsigned short *readval;

    Description: This routine reads a 16-bit value from the
                 register offset specified by addr.

*************************************************************************/

short reada16(unsigned short addr,unsigned short *readval)
{
    unsigned short *ptr = NULL;

    /* create pointer to A16 space of CVME-1149.1 */
    
   sysBusToLocalAdrs(0x29,(char*)addr+a16base,(char **) &ptr);

    *readval = *ptr;            /* read data */

    return 0;
}


/*************************************************************************

       Function: wait_for_ready()

        Summary: Returns when ACT8990 is ready to be accessed

          Usage: void wait_for_ready(void)

    Description: This routine does not return until the ACT8990 *RDY line
                 goes low, indicating the chip is ready to be accessed.
                 This function has no parameters.

**************************************************************************/

void wait_for_ready()
{
    unsigned short status;
    
    do {
        reada16(VME_STATUS, &status);       /* read status bit */
    } while(status & 0x1000);               /* check RDY bit */
}


/*************************************************************************

       Function: check_ready()

        Summary: Returns the value of the ACT8990 *RDY line

          Usage: void check_ready(void)

    Description: This routine returns the value of the ACT8990 *RDY line.
                 This function has no parameters.

**************************************************************************/

short check_ready()
{
    unsigned short status;

    reada16(VME_STATUS, &status);           /* read status bit */

    return((status & 0x1000) == 0);         /* return status of RDY bit */
}


/*************************************************************************

       Function: write_tbc()

        Summary: Writes a value to one of the TBC registers

          Usage: void write_tbc(register_offset, data)
                 unsigned short register_offset;
                 unsigned short data;

    Description: This routine first checks the register_offset to make sure
                 it is a valid offset.  If it is not, the function returns.
                 When the ACT8990 is ready to be accessed, the specified
                 data is written to the register.  This function has two
                 parameters.  The 'register_offset' parameter specifies
                 which of the ACT8990 registers is to be written to.  The
                 'data' parameter holds the data to be written.

**************************************************************************/

void write_tbc(unsigned short register_offset,unsigned short data)
{
    if((register_offset < 0x10) || (register_offset > 0x3e) || 
       (register_offset & 1))
             return;
    else {
        wait_for_ready();               /* wait until ACT8990 is ready */
        writea16(register_offset, data);
    }
}


/*************************************************************************

       Function: read_tbc()

        Summary: Writes a value to one of the TBC registers

          Usage: void write_tbc(register_offset, data)
                 unsigned short register_offset;
                 unsigned short *data;

    Description: This routine first checks the register_offset to make sure
                 it is a valid offset.  If it is not, the function returns.
                 When the ACT8990 is ready to be accessed, the specified
                 register is read and the data is returned.  This function
                 has two parameters.  The 'register_offset' parameter 
                 specifies which ACT8990 register is to be read.  The 'data'
                 parameter contains the address where the read data is to
                 be stored.

**************************************************************************/

void read_tbc(unsigned short register_offset,unsigned short *data)
{
    if((register_offset < 0x10) || (register_offset > 0x3e) ||
       (register_offset & 1))
            return;
    else {
        wait_for_ready();               /* wait for ACT8990 ready */
        reada16(register_offset,data);
    }
}


/*************************************************************************

       Function: write_buf()

        Summary: Writes data to the JTAG data output register

          Usage: void write_buf(data)
                 unsigned short data;

    Description: This routine polls the STATUS2 register until the
                 write_ok flag is clear.  At that time, it writes the
                 specified data to the JTAG write buffer.  This function
                 has one parameter.  The 'data' parameter specifies the
                 data to be written to the JTAG output buffer.

**************************************************************************/

void write_buf(unsigned short data)
{
    /* writes a value to the JTAG data output register */

    unsigned short status;

    do {
        write_tbc(MINOR,0x6040);        /* capture status */

        wait_for_ready();               /* wait for 8990 ready */

        reada16(STATUS2, &status);

    } while(status & 0x0100);           /* write_ok? */

    writea16(WRITEBUF, data);           /* write data to send */
}


/*************************************************************************

       Function: read_buf()

        Summary: Reads data from the JTAG data input buffer

          Usage: void read_buf(data)
                 unsigned short *data;

    Description: This routine polls the STATUS2 register until the
                 read_ok flag is clear.  At that time, it reads the
                 JTAG read buffer, and stores the read data.  This
                 fuction has one parameter.  The 'data' parameter 
                 contains the address where to store the data read
                 from the JTAG input buffer.

**************************************************************************/

void read_buf(unsigned short *data)
{
    unsigned short status;

    /* reads a value from the JTAG data input register */

    do {
        write_tbc(MINOR,0x6040);        /* capture status */

        wait_for_ready();               /* wait for 8990 ready */

        reada16(STATUS2, &status);

    } while(!(status & 0x1000));        /* read_ok? */

    reada16(READBUF, data);             /* read from input buffer */
}

/*************************************************************************

       Function: wait_jtag_state()

        Summary: Waits for the JTAG bus to transition to a particular state

          Usage: void wait_jtag_state(state);
                 unsigned short state;

    Description: This routine polls the STATUS2 register until the
                 JTAG bus is in the specified state.

**************************************************************************/

void wait_jtag_state(short state)
{
    unsigned short status;
    int i=100000;
    
    do {
        write_tbc(MINOR,0x6040);        /* capture status */

        wait_for_ready();               /* wait for 8990 ready */

        reada16(STATUS2,&status);
	
	i--;
	
    } while(((status & 7) != state)&&i);
}

unsigned short jtag_state(void)
{
    unsigned short status;
    
    write_tbc(MINOR,0x6040);        /* capture status */

    wait_for_ready();               /* wait for 8990 ready */

    reada16(STATUS2,&status);
	
    return status;
}


/*************************************************************************

       Function: hard_reset()

        Summary: Provides a hardware reset of the ACT8990

          Usage: short hard_reset(target, divider)
                 unsigned short target;
                 unsigned short divider;

    Description: This routine causes the PC-AT bus board to provide a
                 hardware reset of the 8990 IC by pulsing the 8990
                 *TRST input.  This also pulses the target *TRST signal.
                 There are two inputs to this function.  The 'target'
                 parameters specifies whether the board is to connect
                 the JTAG path to the on-board JTAG 8244 devices or to the
                 external JTAG target.  The second parameter specifies the
                 TCLK divider to use.

                 This function can also be used to verify the existance
                 of the JTAG board in a system.  When values are written
                 to the board, they are read back to verify the board is
                 operating normally.  If the board is responding, this
                 function returns TRUE.  If the board is not responding,
                 this function returns FALSE.  Extensive checking is not
                 done to verify the board responding is indeed a PC-1149.1.

**************************************************************************/

short hard_reset(unsigned short target,unsigned short clock_source,unsigned short divider)
{
    unsigned short x, control;

    /* check values */
    if((target > 1) || (clock_source > 3) || (divider > 15))
        return -1;

    /* setup configuration register 1 value */
    control = 0x0080 |
              ((clock_source & 2) << 11) |
              ((clock_source & 1) << 4) |
              divider | 0xf00;

    if(target == TBC_EXTERNAL)
        control |= 0x20;                    /* external target */

    writea16(CFG_REG1, control);            /* assert reset, tristate target */

    reada16(CFG_REG1, &x);                  /* read-back to verify */
    if(x != control)
        return FALSE;

    control |= 0x0040;                      /* deassert reset */

    writea16(CFG_REG1, control);

    reada16(CFG_REG1, &x);                  /* read-back to verify */
    if(x == control)
        return FALSE;

    writea16(CFG_REG2, 0);              /* disconnect triggers & interrupts */

    reada16(CFG_REG2,&x);                   /* read-back to verify */
    if(x == 0)
        return FALSE;

    return TRUE;
}

/*************************************************************************

       Function: soft_reset()

        Summary: Provides a software reset of the ACT8990

          Usage: void soft_reset(void)

    Description: This routine causes the PC-AT bus board to provide a
                 software reset of the 8990 IC by issuing a MINOR reset
                 command.  This function has no parameters.

**************************************************************************/

void soft_reset()
{
    write_tbc(MINOR,0x7001);        /* soft-reset ACT8990 */
}


/*************************************************************************

       Function: scan_ir()

        Summary: Scans a bit stream into the TAP instruction register

          Usage: void scan_ir(output, length, input)
                 unsigned short *output;
                 unsigned short length;
                 unsigned short *input;

    Description: This routine scans an arbitrary length bit stream into
                 the target TAP controller instruction register.  When a
                 scan starts, the RX-FIFO holds garbage, so we always shift
                 out a 16-bit dummy value first to take out the garbage.
                 This function has three parameters.  The 'output' variable
                 is an address to an array of 16-bit values to be scanned
                 out the JTAG port.  The 'length' parameter specifies the
                 length of the bitstream.  The 'input' variable is an address
                 to an array of 16-bit values where the data scanned in is
                 to be stored.

**************************************************************************/

void scan_ir(unsigned short *output,unsigned short length,unsigned short *input)
{
    unsigned short x,numregs;

    /* determine number of registers to use in data transfer */
    numregs = length / 16;

    if((numregs * 16) != length)
        ++numregs;
    
    write_tbc(CONTROL3,0x008f);		/* start TCK */

    write_tbc(CONTROL4,0x0280+tck_delay);         /* connect TDI0 to data-in-src */
                                        /* send recirculate data from shifter FIFO */
                                        /* ignore received data */
                                        /* FIFO length = 16 */

    write_tbc(MAJOR,0x300f);            /* SCAN command, use counter1 */
                                        /* scan IR, end in pause-ir */

    write_tbc(CNT1UP0,15);              /* counter1 update value = 16 bits */
    write_tbc(MINOR,0x6001);            /* OPERATE2, update counter1 */
        
    write_tbc(MINOR,0x4080);            /* OPERATE0, begin SCAN command */

    wait_jtag_state(PAUSE_IR);          /* wait until SCAN command finishes */
                                        /* and JTAG bus goes to pause-ir */


    write_tbc(CONTROL4,0x0780+tck_delay);         /* connect TDI0 to data-in-src */
                                        /* send overwrite data from write buffer */
                                        /* read receive data into read buffer */
                                        /* FIFO length = 16 */

    write_tbc(MAJOR,0x300a);            /* SCAN command, use counter1 */
                                        /* scan IR, end in run-test-idle */

    write_tbc(CNT1UP0,length - 1);      /* counter1 update value = length - 1 */
    write_tbc(MINOR,0x6001);            /* OPERATE2, update counter1 */

    write_tbc(MINOR,0x4080);            /* OPERATE1, begin SCAN command */

    for(x = 0 ; x < numregs ; ++x) {
        write_buf(output[x]);           /* scan out data */
        read_buf(&input[x]);		/* read data from target */
    }

    wait_jtag_state(RUN_TEST_IDLE);     /* wait until SCAN command finishes */
                                        /* and JTAG bus goes to run-test-idle */
					
    write_tbc(CONTROL3,0x208f);		/* stop TCK */
}


/*************************************************************************

       Function: scan_dr()

        Summary: Scans a bit stream out the TAP data path

          Usage: void scan_dr(output, length, input)
                 unsigned short *output;
                 unsigned short length;
                 unsigned short *input;

    Description: This routine scans an arbitrary length bit stream out
                 the target TAP controller data path.  When a scan starts,
                 the RX-FIFO holds garbage, so we always shift out a 
                 16-bit dummy value first to take out the garbage.
                 This function has three parameters.  The 'output' variable
                 is an address to an array of 16-bit values to be scanned
                 out the JTAG port.  The 'length' parameter specifies the
                 length of the bitstream.  The 'input' variable is an address
                 to an array of 16-bit values where the data scanned in is
                 to be stored.

**************************************************************************/

void scan_dr(unsigned short *output,unsigned short length,unsigned short *input)
{
    unsigned short x,numregs;

    /* this routine scans data thru the JTAG data path */

    /* determine number of registers to use in data transfer */
    numregs = length / 16;

    if((numregs * 16) != length)
        ++numregs;
    write_tbc(CONTROL3,0x008f);		/* start TCK */

    write_tbc(CONTROL4,0x0280+tck_delay);         /* connect TDI0 to data-in-src */
                                        /* send recirculate data from shifter FIFO */
                                        /* ignore received data */
                                        /* FIFO length = 16 */

    write_tbc(MAJOR,0x3006);            /* SCAN command, use counter1 */
                                        /* scan DR, end in pause-dr */

    write_tbc(CNT1UP0,15);              /* counter1 update value = 16 bits */
    write_tbc(MINOR,0x6001);            /* OPERATE2, update counter1 */

    write_tbc(MINOR,0x4080);            /* OPERATE0, begin SCAN command */

    wait_jtag_state(PAUSE_DR);          /* wait until SCAN command finishes */
                                        /* and JTAG bus goes to pause-dr */


    write_tbc(CONTROL4,0x0780+tck_delay);         /* connect TDI0 to data-in-src */
                                        /* send overwrite data from write buffer */
                                        /* read receive data into read buffer */
                                        /* FIFO length = 16 */

    write_tbc(MAJOR,0x3002);            /* SCAN command, use counter1 */
                                        /* scan DR, end in run-test-idle */

    write_tbc(CNT1UP0,length - 1);      /* counter1 update value = length - 1 */
    write_tbc(MINOR,0x6001);            /* OPERATE2, update counter1 */

    write_tbc(MINOR,0x4080);            /* OPERATE1, begin SCAN command */

    for(x = 0 ; x < numregs ; ++x) {
        write_buf(output[x]);           /* scan out data */
        read_buf(&input[x]);		/* read data from target */
    }

    wait_jtag_state(RUN_TEST_IDLE);     /* wait until SCAN command finishes */
                                        /* and JTAG bus goes to run-test-idle */
    write_tbc(CONTROL3,0x208f);		/* stop TCK */
}


/*************************************************************************

       Function: circulate_dr()

        Summary: Circulates a bit stream thru the TAP controller data path

          Usage: void circulate_dr(length, data)
                 unsigned short length;
                 unsigned short *data;

    Description: This routine circulates an arbitrary length bit stream thru
                 the target TAP controller data path.  Data doesnt actually
                 start circulating until the first 32 bits of data are sent
                 out.  The first 16 bits are sent to get valid data into the
                 RX-FIFO, and the second 16 bits force the first valid data
                 out of the scan path.  This function has two parameters.
                 The 'length' parameter specifies how many bits to 
                 circulate.  The 'data' parameter is an address to an array
                 of 16-bit values where the data scanned in is to be stored.

**************************************************************************/

void circulate_dr(unsigned short length,unsigned short *output)
{
    unsigned short x,numregs;

    /* determine number of registers to use in data transfer */
    numregs = length / 16;

    if((numregs * 16) != length)
        ++numregs;
    write_tbc(CONTROL3,0x008f);		/* start TCK */

    write_tbc(CONTROL4,0x0280+tck_delay);         /* connect TDI0 to data-in-src */
                                        /* send recirculate data from shifter FIFO */
                                        /* ignore receive data */
                                        /* FIFO length = 16 */

    write_tbc(MAJOR,0x3006);            /* SCAN, use counter1 */
                                        /* scan DR, end in pause-dr */

    write_tbc(CNT1UP0,15);              /* counter1 update value = 16 bits */
    write_tbc(MINOR,0x6001);            /* OPERATE2, update counter1 */

    write_tbc(MINOR,0x4080);            /* OPERATE0, begin SCAN command */

    wait_jtag_state(PAUSE_DR);          /* wait until SCAN command finishes */
                                        /* and JTAG bus goes to pause-dr */


    write_tbc(CONTROL4,0x0780+tck_delay);         /* connect TDI0 to data-in-src */
                                        /* send overwrite data from write buffer */
                                        /* read receive data into read buffer */
                                        /* FIFO length = 16 */

    write_tbc(MAJOR,0x3002);            /* SCAN command, use counter1 */
                                        /* scan DR, end in run-test-idle */

    write_tbc(CNT1UP0,length - 1 + 16); /* counter1 update value=length-1+16 */
    write_tbc(MINOR,0x6001);            /* OPERATE2, update counter1 */

    write_buf(0);                       /* scan dummy value to get first data */
                                        /* from target for recirculation */

    write_tbc(MINOR,0x4080);            /* OPERATE0, begin SCAN command */

    /* read and recirculate data into the target */
    for(x = 0 ; x < numregs ; ++x) {
        read_buf(&output[x]);           /* read result of DR-scan */
        write_buf(output[x]);           /* rescan data into target */
    }

    read_buf(&x);                       /* read last word which is garbage */

    wait_jtag_state(RUN_TEST_IDLE);     /* wait until SCAN command finishes */
                                        /* and JTAG bus goes to run-test-idle */
    write_tbc(CONTROL3,0x208f);		/* stop TCK */
}

short reset_jtag()
{
    short err;

    err=hard_reset(TBC_EXTERNAL,CLK_781KHZ,TBC_DIV_BY_4);
    write_tbc(CONTROL0,0);
    write_tbc(CONTROL1,0);
    write_tbc(CONTROL2,1);
    write_tbc(CONTROL3,0x8f);
    write_tbc(CONTROL5,0);
    write_tbc(CONTROL6,0);
    write_tbc(CONTROL7,0);
    write_tbc(CONTROL8,4);
    write_tbc(CONTROL9,0);
    return err;
    
}

short reset_jtag_var(unsigned short i,unsigned short j,unsigned short k)
{
    short err;

    err=hard_reset(i,j,k);
    write_tbc(CONTROL0,0);
    write_tbc(CONTROL1,0);
    write_tbc(CONTROL2,1);
    write_tbc(CONTROL3,0x8f);
    write_tbc(CONTROL5,0);
    write_tbc(CONTROL6,0);
    write_tbc(CONTROL7,0);
    write_tbc(CONTROL8,4);
    write_tbc(CONTROL9,0);
    return err;
    
}

void connect_tap(char tap)
{
   writea16(PIO1_OUT,tap);

   if(tap<64) write_tbc(CONTROL2,1<<((tap&0x30)>>4));
}

void disconnect_tap(char tap)
{
   unsigned short ctrl2;

   read_tbc(CONTROL2,&ctrl2);
   ctrl2&=~(1<<tap);
   write_tbc(CONTROL2,ctrl2);
}


/***************************************************************************************


	bit_to_word (bit_nb)
	
	unsigned short	bit_nb
	
	return : unsigned short word_nb
	
	
	Description : 	this function determines how many words are needed to represent
			bit_nb bits


****************************************************************************************/
unsigned short bit_to_word(unsigned short bit_nb)
{
    unsigned short word_nb;
    
    word_nb=bit_nb/16;
    if ((word_nb*16) == bit_nb) word_nb--;
    
    return word_nb;
}

/***************************************************************************************


	map_nb(map)
	
	chip 		*map
	
	return : unsigned short chip_nb
	
	Description :  	this function determines how many chip are in the chain, 
			according to the "map".


****************************************************************************************/

unsigned short map_nb(chip *map)
{
    unsigned short i=0;
    
    for (;(map[i]!=END)&&(i<MAX_CHIPS);i++);
    if (i==MAX_CHIPS) i=0;
    return i;
}


/***************************************************************************************


	init_stream(stream,length)
	
	unsigned short 	*stream
	unsigned short 	length
	
	return :	nothing
	
	Description : this function sets the stream to be all ones
		     (all ones instruction = BYPASS instruction, according to IEEE 1149.1)	


******************************************************************************************/

void init_stream(unsigned short *stream,unsigned short length)
{
   
   int i,n;

   n=bit_to_word(length);      
   for (i=0;i<(n+1);i++)
       stream[i]=0xffff;
       
}

/***************************************************************************************


	set_stream(stream,loc_bit,data,data_sz)
	
	unsigned short 	*stream
	unsigned short 	loc_bit
	unsigned long 	data
	unsigned short 	data_sz
	
	return :	nothing
	
	
	Description : 	this function puts "data" in the stream at the location 
			corresponding to "loc_bit".
				

******************************************************************************************/

void set_stream(unsigned short *stream,unsigned short loc_bit,unsigned long data,unsigned short data_sz)
{

unsigned short w1,w2,b1,b2,data_mask,i;
    
    w1 = loc_bit/16;		/* nb of word where data begins 			*/
    b1 = loc_bit - w1*16;		/* nb of bit where data begins in w1 			*/
    w2 = bit_to_word(loc_bit+data_sz);	/* nb of word where data ends 				*/			
    b2 = (loc_bit+data_sz) - w2*16;	/* nb of bit where data ends in w2			*/
    
   /* printf("set stream : str:%p loc:%d data:%x  siz:%d\n",stream,loc_bit,data,data_sz); */

    if (w1==w2)  /* if the data fits in a single word of the stream */
    {
       for (data_mask=0,i=0;i<16;i++)
	  if ((i<b1)||(i>=b2)) data_mask+=1<<i;
       stream[w1]=(stream[w1]&data_mask)+(data<<b1);	/* fill this single word */
       return;
    }
    else
    {
       for (data_mask=0,i=0;i<16;i++)
          if (i<b1) data_mask+=1<<i;
       stream[w1]=(stream[w1]&data_mask)+((data<<b1)&0xffff);	/* fill the first word */
       
       for (data_mask=0,i=0;i<16;i++)
          if (i>=b2) data_mask+=1<<i;
       stream[w2]=(stream[w2]&data_mask)+(data>>(data_sz-b2));	/* fill the last word */
       
       if ((w1+1)<w2) 	/* if the data fits in three words of the stream */
          stream[w1+1]=(data>>(16-b1))&0xffff;			/* fill the middle word */
    }
       
}  
   
/***************************************************************************************


	get_stream(stream,loc_bit,data_sz)
	
	unsigned short 	*stream
	unsigned short 	loc_bit
	unsigned short 	data_sz
	
	return :	unsigned long data
	
	
	Description : 	this function gets the data from the stream at the location 
			corresponding to "loc_bit" (data_sz : number of bits of the data)
				

******************************************************************************************/

unsigned long	get_stream(unsigned short *stream,unsigned short loc_bit,unsigned short data_sz)
{

unsigned short w1,w2,b1,b2,data_mask,i;
unsigned long data;


    /*printf("get stream : str:%p loc:%d siz:%d\n",stream,loc_bit,data_sz);*/

    w1 = loc_bit/16;		/* nb of word where data begins 			*/
    b1 = loc_bit - w1*16;		/* nb of bit where data begins in w1 			*/
    w2 = bit_to_word(loc_bit+data_sz);	/* nb of word where data ends 				*/
    b2 = (loc_bit+data_sz) - w2*16;	/* nb of bit where data ends in w2			*/
    
   /* printf("get stream : w1:%d w2:%d b1:%d b2:%d\n",w1,w2,b1,b2); */
    
    if (w1==w2)  /* if the data fits in a single word of the stream */
    {
       for (data_mask=0,i=0;i<16;i++)
	  if ((i>=b1)&&(i<b2)) data_mask+=1<<i;
       data=(stream[w1]&data_mask)>>b1;		/* get the data */ 
       return data;
    }
    else
    {
       for (data_mask=0,i=0;i<16;i++)
          if (i>=b1) data_mask+=1<<i;
       data=(stream[w1]&data_mask)>>b1;	/* get the part of the data from the first word */
  
       /*printf("get stream : data1:%x",(stream[w1]&data_mask)>>b1);*/
           
       for (data_mask=0,i=0;i<16;i++)
          if (i<b2) data_mask+=1<<i;
       data+=((unsigned long)(stream[w2]&data_mask))<<((w2-w1)*16-b1);	/* get the part of the data from the last word */
       /*printf(" data2:%x",((unsigned long)(stream[w2]&data_mask))<<((w2-w1)*16-b1));*/
       
       if ((w1+1)<w2) 	/* if the data fits in three words of the stream */
          {data+=((unsigned long)stream[w1+1])<<(16-b1);			/* get the part of the data from the middle word */
       /*printf(" data3:%x\n",((unsigned long)stream[w1+1])<<(16-b1));*/
    	}
    }
    return data;
}

/***************************************************************************************

	bypass_all_chips(map)
	
	Description : 	this function send a bypass instruction to all the chips 
			(map descibes the chain).
				

******************************************************************************************/
void bypass_all_chips(chip *map)
{
    unsigned short stream_in[MAX_LENGTH_STREAM_IR],stream_out[MAX_LENGTH_STREAM_IR];
    unsigned short nb_chip;
    
    nb_chip = map_nb(map);
    
    init_stream(stream_out,IR_SZ*nb_chip);	/* fill the output stream with BYPASS instructions */
    
    scan_ir(stream_out,IR_SZ*nb_chip,stream_in);
    
}

/***************************************************************************************

	jtag_check()
	
	return : int
		
	Description : 	this function check the continuity of the jtag chain and return 1
			if the chain is not interrupted.
						

******************************************************************************************/ 
int jtag_check(unsigned short ir)
{
   unsigned short stream_in[MAX_LENGTH_STREAM_IR],stream_out[MAX_LENGTH_STREAM_IR],i;


   /* for(i=0;i<MAX_LENGTH_STREAM_IR;i++) stream_out[i]=0xffff;*/
   stream_out[0]=ir;
   scan_ir(stream_out,IR_SZ,stream_in); /* bypass all chips */
   
   
   for(i=0;i<MAX_LENGTH_STREAM_IR;i++) stream_out[i]=0x5555;
   scan_dr(stream_out,(MAX_LENGTH_STREAM_IR-2)*16,stream_in); 	/* fill the bypass registers with 010101... */

   for(i=0;i<MAX_LENGTH_STREAM_IR;i++) printf("%4x ",stream_in[i]);
   return(stream_in[MAX_CHIPS/16+2]==0xaaaa);
}
   
   
   
/***************************************************************************************

	detect_length()
	
	return : unsigned short length
		
	Description : 	this function detect automatically the length of the JTAG chain and 
			return the number of chips in the chain.
			
	note : if this function is not called just after a reset, it should be called twice. 
			

******************************************************************************************/ 
unsigned short detect_length()
{
   unsigned short stream_in[MAX_LENGTH_STREAM_IR],stream_out[MAX_LENGTH_STREAM_IR],i;


   for(i=0;i<MAX_LENGTH_STREAM_IR;i++) stream_out[i]=0xffff;	
   scan_ir(stream_out,IR_SZ*MAX_CHIPS+16,stream_in); /* bypass all chips */
   
   
   for(i=0;i<MAX_LENGTH_STREAM_IR;i++) stream_out[i]=0;
   scan_dr(stream_out,MAX_CHIPS+16,stream_in); 	/* fill the bypass registers with zeros */

   
   i=0;
   while((get_stream(stream_in,i,1)==0)&&(i<MAX_CHIPS)) i++;	
   						/* count the zeros (it works because the FIFO of the TBC was filled with
   						   ones because of the scan_ir command and the bypass registers were filled 
						   with zeros because of the reset done before this function is called )   */
   if (i==MAX_CHIPS) i=0;
   
   return i;
}
   
void clear_bypass()
{
   unsigned short stream_in[MAX_LENGTH_STREAM_IR],stream_out[MAX_LENGTH_STREAM_IR],i;

   for(i=0;i<MAX_LENGTH_STREAM_IR;i++) stream_out[i]=0xffff;	
   scan_ir(stream_out,IR_SZ*MAX_CHIPS+16,stream_in); /* bypass all chips */
   
   
   for(i=0;i<MAX_LENGTH_STREAM_IR;i++) stream_out[i]=0;
   scan_dr(stream_out,MAX_CHIPS+16,stream_in); 	/* fill the bypass registers with zeros */

}
   
/***************************************************************************************

	auto_init_map(map)
	
	chip *map
	
	return : number of chips
	
	Description : 	this function initialize the chip map automatically by sending EXTEST
			commands to distinguish between COSTAR and ALICE128C chips.
			

******************************************************************************************/ 

int auto_init_map(chip *map)
{
   unsigned short nb_chip; 
   unsigned short stream_in[MAX_LENGTH_STREAM_IR],stream_out[MAX_LENGTH_STREAM_IR];
   int i;
      
   if ((nb_chip=detect_length())==0) {map[0]=END; return 0;}
   
   for (i=0;i<nb_chip;i++)
   {
       init_stream(stream_out,IR_SZ*nb_chip);
       
       set_stream(stream_out,i*IR_SZ,0x1b,IR_SZ);   
   
       scan_ir(stream_out,IR_SZ*nb_chip,stream_in); 	/* send ID command */
   
       circulate_dr(nb_chip*BYPASS_SZ+IDENT_SZ-1,stream_in);	/* read the ID register (result in stream_in) */			
     /*printf("%d ",get_stream(stream_in,i*BYPASS_SZ,8));*/
     
      if (get_stream(stream_in,i*BYPASS_SZ,8)==0xaf)	     /* if the ID register is the costar one */
      {
         map[i]=COSTAR;					/* it's a COSTAR */
      }
      else if ((get_stream(stream_in,i*BYPASS_SZ,8)&0xf0)==0x90)	     /* if the ID register is the FPGA connexion one */
      {
         map[i]=CONNEXION;					/* it's a FPGA connexion */
      }
      else if ((get_stream(stream_in,i*BYPASS_SZ,8)&0xf0)==0x60)	     /* if the ID register is the FPGA readout one */
      {
         map[i]=READOUT;					/* it's a FPGA readout */
      }
      else
      {
         map[i]=ALICE128C;				/* else it's a ALICE128C */
      }
   }
   
   map[nb_chip]=END;
   return nb_chip;
}
       

   
/***************************************************************************************

	init_map(map,name)
	
	chip *map
	char *name
	
	return : nothing
	
	Description : 	this function initialize the chip map by reading a ascii file which
			name is in the string 'name'. 
			
			
	note : 		in the ascii file :
			
			'a' or 'A' means Alice128C chip
			'c' or 'C' means Costar chip
			'.' means end of description.

******************************************************************************************/ 

void init_map(chip *map,char *name)
{
   FILE *map_file;
   int i,c,type;
   
   map[0]=COSTAR;
   type=0;
   for(i=0;name[i]!='\0';i++)
      if ((name[i]!='c')&&(name[i]!='a')&&(name[i]!='C')&&(name[i]!='A')&&(name[i]!='.')&&(name[i]!='x')&&(name[i]!='X'))
         type=1;
   if (type==1)
   {
      name=strcat(name,".map");
      map_file = fopen(name,"r");
      if (map_file==NULL)
        {
	   printf("\nERROR : can't open map file.");
	   map[0]=END;
	   return;
 	}
      fgets(name,128,map_file);
      fclose(map_file);
   }
   i=0;
   for(i=0;name[i]!='\0';i++)
   {
      switch(c=name[i])
      {
	 case '.' : map[i]=END;
	 	    break;
		    
	 case 'c' : map[i]=COSTAR;
	 	    break;
		    
	 case 'C' : map[i]=COSTAR;
	 	    break;
		    
	 case 'a' : map[i]=ALICE128C;
	 	    break;
		    
	 case 'A' : map[i]=ALICE128C;
	 	    break;
		    
	 case 'x' : map[i]=CONNEXION;
	 	    break;
		    
	 case 'X' : map[i]=CONNEXION;
	 	    break;
		    
	 case 'r' : map[i]=READOUT;
	 	    break;
		    
	 case 'R' : map[i]=READOUT;
	 	    break;
		    
	 default  : break;
       }
     
    }
    if((i>0)&&(map[i-1]!=END)) map[i]=END;
    if(!i) map[i]=END;
     
    
    return;
}

/***************************************************************************************

	jtag_init(map,chain)
	
	chip *map
	char *chain
	
	return : nothing
	
	Description : 	this function try to initialize the jtag map (automatically) and
			put it in "map". If it doesn't work, it uses the default map
			described by "chain". "chain" is the name of the ascii file 
			describing the map, or the map itself.
	
	ex :		jtag_init(map,"default")  use the file "default.map" as default map
			jtag_init(map,"aaaaaac.") the default map is 6 A128C and 1 COSTAR.  
			
			
	note : 		in the ascii file :
			
			'a' or 'A' means Alice128C chip
			'c' or 'C' means Costar chip
			'.' means end of description.
			
			the name of this file must contain at least one character different
			from 'a' or 'c'

******************************************************************************************/ 
   
int jtag_init(chip *map,char *chain)
{
   
   detect_length();
   if (!auto_init_map(map))
      {
      printf(" ERROR : JTAG chain not connected.\nUsing default map\n");
      init_map(map,chain);
      }
   if (map[0]==END) return 1;
   else return 0;
}   
    
void uncompress_map(chip *map,char *cmap)
{
   int i,j,fin=0;
   
   for (i=0;i<COMPRESSED_MAP_LENGTH;i++)
   {
      for (j=0;j<4;j++)
         if ((map[i*4+j]=(cmap[i]&(3<<(2*j)))>>(2*j))==END) 
	 {
	    fin=1;
	    break;
	 }
   if (fin) break;
   }
   
}

void compress_map(chip *map,char *cmap)
{
    int i,j,fin=0;
    chip c;
    
   for (i=0;i<COMPRESSED_MAP_LENGTH;i++)
   {
      for (j=0,cmap[i]=0;j<4;j++)
      {
         c=map[i*4+j];
	 if (c>END) c=ALICE128C;
	 cmap[i]|=(c&3)<<(2*j);
	 if (c==END)
	 {
	    fin=1;
	    break;
	 }
      }
   if(fin) break;
   }
}

void ascii_map(chip *map,char *amap)
{
    int i;
    
   for (i=0;(i<MAX_CHIPS)&&(map[i]!=END);i++)
   {
      switch(map[i])
      {
      case COSTAR    : amap[i]='C';
      		       break;
      case ALICE128C : amap[i]='A';
      		       break;
      case CONNEXION : amap[i]='X';
      		       break;
      case READOUT   : amap[i]='R';
      		       break;
      default        : break;
      }
   }	    
   amap[i++]='.';
   amap[i]=0;
}
