/* jtagConnexion.c		Jtag FPGA connexion functions 
 *
 *	Institut de Recherche Subatomique de Strasbourg
 *
 *	Author :	David Bonnet
 *	Date :		31/05/00
 *
 */

 
#include "vxWorks.h"
#include "sysLib.h"
#include "stdio.h"
#include "jtagCommon.h"
#include "jtagConnexion.h"

 /***************************************************************************************


	map_connexion_nb(map)

	chip 		*map

	return : 	unsigned short connexion_nb

	Description :  	this function determines how many FPGA connexion are in the chain,
			according to the "map".


****************************************************************************************/

unsigned short map_connexion_nb(chip *map)
{
    unsigned short i=0,iconn=0;

    for (;(map[i]!=END)&&(i<MAX_CHIPS);i++)
        if (map[i]==CONNEXION) iconn++;

    return iconn;
}


/***************************************************************************************

	send_one_connexion(tap,instruction,instr_siz,data,connexion_nb,map)

    	int		tap
	unsigned long 	instruction
	unsigned short 	instr_siz
	unsigned long 	data
	unsigned short 	connexion_nb
	chip 		*map

	return : 	unsigned long old_data


	Description : 	this function send the instruction to the FPGA connexion "connexion_nb"
			(map descibes the chain), then send the data. the data returned
			by the FPGA connexion (the old data) is returned by this function.


******************************************************************************************/


unsigned long send_one_connexion(unsigned long instruction,unsigned short instr_siz,
			      unsigned long data,unsigned short connexion_nb,chip *map)
{
    unsigned short stream_in[MAX_LENGTH_STREAM],stream_out[MAX_LENGTH_STREAM];
    unsigned short nb_chip,i,iconn,conn_i=0;


    if((map_connexion_nb(map))==0) return 0;
    nb_chip = map_nb(map);
    for (i=0,iconn=0;i<nb_chip;i++)            /* find the connexion in the chain */
        if (map[i]==CONNEXION)
	{
	    if (iconn==connexion_nb)
	    {
	        conn_i=i;
                break;
	    }
	    else iconn++;
	}

    init_stream(stream_out,(unsigned short)(CX_IR_SZ*nb_chip));	/* fill the output stream with BYPASS instructions */

    set_stream(stream_out,(unsigned short)(conn_i*CX_IR_SZ),instruction,CX_IR_SZ); /* put the instruction for the connexion in the output stream. */

    scan_ir(stream_out,(unsigned short)(CX_IR_SZ*nb_chip),stream_in);		/* send the instruction stream */

    set_stream(stream_out,(unsigned short)(conn_i*CX_BYPASS_SZ),data,instr_siz);	/* create the data stream to be sent */

    scan_dr(stream_out,(unsigned short)(nb_chip*CX_BYPASS_SZ+instr_siz-1),stream_in); /* send the data stream */


    return get_stream(stream_in,(unsigned short)(conn_i*CX_BYPASS_SZ),instr_siz);   /* return the old data (returned by the chip) */
}

/***************************************************************************************

	get_one_connexion(tap,instruction,instr_siz,connexion_nb,map)

    	int		tap
	unsigned long 	instruction
	unsigned short 	instr_siz
	unsigned short 	connexion_nb
	chip 		*map

	return : 	unsigned long old_data


	Description : 	this function send the instruction to the FPGA connexion "connexion_nb"
			(map descibes the chain), then send the data. the data returned
			by the FPGA connexion (the old data) is returned by this function.


******************************************************************************************/


unsigned long get_one_connexion(unsigned long instruction,unsigned short instr_siz,
			      unsigned short connexion_nb,chip *map)
{
    unsigned short stream_in[MAX_LENGTH_STREAM],stream_out[MAX_LENGTH_STREAM];
    unsigned short nb_chip,i,iconn,conn_i=0;


    if((map_connexion_nb(map))==0) return 0;
    nb_chip = map_nb(map);
    for (i=0,iconn=0;i<nb_chip;i++)            /* find the connexion in the chain */
        if (map[i]==CONNEXION)
	{
	    if (iconn==connexion_nb)
	    {
	        conn_i=i;
                break;
	    }
	    else iconn++;
	}

    init_stream(stream_out,(unsigned short)(CX_IR_SZ*nb_chip));	/* fill the output stream with BYPASS instructions */

    set_stream(stream_out,(unsigned short)(conn_i*CX_IR_SZ),instruction,CX_IR_SZ); /* put the instruction for the connexion in the output stream. */

    scan_ir(stream_out,(unsigned short)(CX_IR_SZ*nb_chip),stream_in);		/* send the instruction stream */

    circulate_dr((unsigned short)(nb_chip*CX_BYPASS_SZ+instr_siz-1),stream_in); /* send the data stream */


    return get_stream(stream_in,(unsigned short)(conn_i*CX_BYPASS_SZ),instr_siz);   /* return the old data (returned by the chip) */
}


/***************************************************************************************

	get_power_status_connexion(tap,connexion_nb,map)

	chip  		map[number of chip]

	return : unsigned short power_status

	Description : 	this function gets the power status register of the FPGA connexion "connexion_nb" 
			and returns it (map describes the chain).


******************************************************************************************/
unsigned short get_power_status_connexion(chip *map)
{
   return (unsigned short)get_one_connexion(CX_STATUS,CX_STATUS_SZ,0,map);
}

/***************************************************************************************

	set_power_connexion(tap,data,connexion_nb,map)

	unsigned short  data
	chip  		map[number of chip]

	return : unsigned short power

	Description : 	this function allows to power on the hybrids which have been triped by
			a latchup and returns the last value (map describes the chain).


******************************************************************************************/
unsigned short set_power_connexion(unsigned short data,chip *map)
{
   return (unsigned short)send_one_connexion(CX_POWER,CX_POWER_SZ,data,0,map);
}

/***************************************************************************************

	get_latchup_connexion(tap,connexion_nb,map)

	chip  		map[number of chip]

	return : unsigned char latchup

	Description : 	this function gets the latchup status register of the FPGA connexion "connexion_nb" 
			and returns it (map describes the chain).


******************************************************************************************/
unsigned char get_latchup_connexion(chip *map)
{
   return (unsigned char)get_one_connexion(CX_LATCHUP,CX_LATCHUP_SZ,0,map);
}



unsigned char get_id_connexion(chip *map)
{
   return (unsigned char)get_one_connexion(CX_IDENT,CX_IDENT_SZ,0,map);
}

/***************************************************************************************

	set_vref_connexion(tap,data,connexion_nb,map)

	unsigned char   data
	chip  		map[number of chip]

	return : unsigned char vref

	Description : 	this function sets the voltage reference for the latchup detection of the
			FPGA connexion "connexion_nb"and returns the last value. 
			(map describes the chain).


******************************************************************************************/
unsigned char set_vref_connexion(unsigned char data,chip *map)
{
   return (unsigned char)send_one_connexion(CX_VREF,CX_VREF_SZ,data,0,map);
}

/***************************************************************************************

	set_bypass_connexion(tap,data,connexion_nb,map)

	unsigned short  data
	chip  		map[number of chip]

	return : unsigned short hybrid_bypass

	Description : 	this function sets the hybrid bypass register of the
			FPGA connexion "connexion_nb" and returns the last value. 
			(map describes the chain).


******************************************************************************************/
unsigned short set_bypass_connexion(unsigned short data,chip *map)
{
   return (unsigned short)send_one_connexion(CX_HYB_BYP,CX_HYB_BYP_SZ,data,0,map);
}

/***************************************************************************************

	set_input_mux_connexion(data,map)

	unsigned short  data[number of modules(16)]
	chip  		map[number of chip]

	return : nothing

	Description : 	this function sets the input mux register of the
			FPGA connexion "connexion_nb". 
			(map describes the chain).
			data in an array of 16 words of ten bits

******************************************************************************************/
void set_input_mux_connexion(unsigned short *data,chip *map)
{
    unsigned short stream_in[MAX_LENGTH_STREAM],stream_out[MAX_LENGTH_STREAM];
    unsigned short nb_chip,i,conn_i=0;


    if((map_connexion_nb(map))==0) return ;
    nb_chip = map_nb(map);
    for (i=0;i<nb_chip;i++)            /* find the connexion in the chain */
        if (map[i]==CONNEXION)
	{
	   conn_i=i;
           break;
	}

    init_stream(stream_out,(unsigned short)(CX_IR_SZ*nb_chip));	/* fill the output stream with BYPASS instructions */

    set_stream(stream_out,(unsigned short)(conn_i*CX_IR_SZ),CX_MUX,CX_IR_SZ); /* put the instruction for the connexion in the output stream. */

    scan_ir(stream_out,(unsigned short)(CX_IR_SZ*nb_chip),stream_in);		/* send the instruction stream */

    for(i=0;i<MODULE_NB;i++)  
      set_stream(stream_out,(unsigned short)(conn_i*CX_BYPASS_SZ+10*(MODULE_NB-1-i)),data[i],10);	/* create the data stream to be sent */
      /* set_stream(stream_out,(unsigned short)(conn_i*CX_BYPASS_SZ+10*(i)),data[i],10);*/
    
      /* printf("mux:");for(i=0;i<10;i++) printf("%04x",stream_out[9-i]);printf("\n");*/
    scan_dr(stream_out,(unsigned short)(nb_chip*CX_BYPASS_SZ+CX_MUX_SZ-1),stream_in); /* send the data stream */


    return;
}



/***************************************************************************************

	get_temperatures_connexion(temp[3],map)

	unsigned short  temp[3]
	chip  		map[number of chip]

	return : nothing

	Description : 	this function get the temperatures of the
			FPGA connexion. 
			(map describes the chain).
			temp in an array of 3 words of 12 bits

******************************************************************************************/
void get_temperature_connexion(unsigned short temp[3],chip *map)
{
    unsigned short stream_in[MAX_LENGTH_STREAM],stream_out[MAX_LENGTH_STREAM];
    unsigned short nb_chip,i,conn_i=0,timeout=1000;

    if((map_connexion_nb(map))==0) return ;

    nb_chip = map_nb(map);
    for (i=0;i<nb_chip;i++)            /* find the connexion in the chain */
        if (map[i]==CONNEXION)
	{
	   conn_i=i;
           break;
	}

    send_one_connexion(CX_TEMP_START,CX_TEMP_START_SZ,0,0,map);
    while(!send_one_connexion(CX_TEMP_START,CX_TEMP_START_SZ,1,0,map)&&timeout--);
    if(!timeout){
      for(i=0;i<3;i++)
	temp[i]=2730;
      return;
    }
    init_stream(stream_out,(unsigned short)(CX_IR_SZ*nb_chip));	/* fill the output stream with BYPASS instructions */
    set_stream(stream_out,(unsigned short)(conn_i*CX_IR_SZ),CX_TEMP,CX_IR_SZ); /* put the instruction for the connexion in the output stream. */
    scan_ir(stream_out,(unsigned short)(CX_IR_SZ*nb_chip),stream_in);		/* send the instruction stream */
    circulate_dr((unsigned short)(nb_chip*CX_BYPASS_SZ+CX_TEMP_SZ-1),stream_in); /* send the data stream */
    for(i=0;i<3;i++)
      temp[i]=get_stream(stream_in,(unsigned short)(conn_i*CX_BYPASS_SZ)+i*(CX_TEMP_SZ/3),CX_TEMP_SZ/3);
}
