/* jtagReadout.c		Jtag FPGA Readout Board functions 
 *
 *	Institut de Recherche Subatomique de Strasbourg
 *
 *	Author :	David Bonnet
 *	Date :		18/06/01
 *
 */


#include "vxWorks.h"
#include "sysLib.h"
#include "stdio.h"
#include "jtagCommon.h"
#include "jtagReadout.h"

 /***************************************************************************************


	map_redout_nb(map)

	chip 		*map

	return : 	unsigned short readout_nb

	Description :  	this function determines how many FPGA readout are in the chain,
			according to the "map".


****************************************************************************************/

unsigned short map_readout_nb(chip *map)
{
    unsigned short i=0,iread=0;

    for (;(map[i]!=END)&&(i<MAX_CHIPS);i++)
        if (map[i]==READOUT) iread++;

    return iread;
}


/***************************************************************************************

	send_one_readout(instruction,instr_siz,data,readout_nb,map)

	unsigned long 	instruction
	unsigned short 	instr_siz
	unsigned long 	data
	unsigned short 	readout_nb
	chip 		*map

	return : 	unsigned long old_data


	Description : 	this function send the instruction to the FPGA readout "readout_nb"
			(map descibes the chain), then send the data. the data returned
			by the FPGA readout (the old data) is returned by this function.


******************************************************************************************/


unsigned long send_one_readout(unsigned long instruction,unsigned short instr_siz,
			      unsigned long data,unsigned short readout_nb,chip *map)
{
    unsigned short stream_in[MAX_LENGTH_STREAM],stream_out[MAX_LENGTH_STREAM];
    unsigned short nb_chip,i,iread,read_i=0;


    if((map_readout_nb(map))==0) return 0;
    nb_chip = map_nb(map);
    for (i=0,iread=0;i<nb_chip;i++)            /* find the readout in the chain */
        if (map[i]==READOUT)
	{
	    if (iread==readout_nb)
	    {
	        read_i=i;
                break;
	    }
	    else iread++;
	}

    init_stream(stream_out,(unsigned short)(RD_IR_SZ*nb_chip));	/* fill the output stream with BYPASS instructions */

    set_stream(stream_out,(unsigned short)(read_i*RD_IR_SZ),instruction,RD_IR_SZ); /* put the instruction for the connexion in the output stream. */

    scan_ir(stream_out,(unsigned short)(RD_IR_SZ*nb_chip),stream_in);		/* send the instruction stream */

    set_stream(stream_out,(unsigned short)(read_i*RD_BYPASS_SZ),data,instr_siz);	/* create the data stream to be sent */

    scan_dr(stream_out,(unsigned short)(nb_chip*RD_BYPASS_SZ+instr_siz-1),stream_in); /* send the data stream */


    return get_stream(stream_in,(unsigned short)(read_i*RD_BYPASS_SZ),instr_siz);   /* return the old data (returned by the chip) */
}

/***************************************************************************************

	get_one_readout(instruction,instr_siz,readout_nb,map)

	unsigned long 	instruction
	unsigned short 	instr_siz
	unsigned short 	readout_nb
	chip 		*map

	return : 	unsigned long old_data


	Description : 	this function send the instruction to the FPGA readout "readout_nb"
			(map descibes the chain), then send the data. the data returned
			by the FPGA readout (the old data) is returned by this function.


******************************************************************************************/


unsigned long get_one_readout(unsigned long instruction,unsigned short instr_siz,
			      unsigned short readout_nb,chip *map)
{
    unsigned short stream_in[MAX_LENGTH_STREAM],stream_out[MAX_LENGTH_STREAM];
    unsigned short nb_chip,i,iread,read_i=0;


    if((map_readout_nb(map))==0) return 0;
    nb_chip = map_nb(map);
    for (i=0,iread=0;i<nb_chip;i++)            /* find the readout board in the chain */
        if (map[i]==READOUT)
	{
	    if (iread==readout_nb)
	    {
	        read_i=i;
                break;
	    }
	    else iread++;
	}

    init_stream(stream_out,(unsigned short)(RD_IR_SZ*nb_chip));	/* fill the output stream with BYPASS instructions */

    set_stream(stream_out,(unsigned short)(read_i*RD_IR_SZ),instruction,RD_IR_SZ); /* put the instruction for the readout in the output stream. */

    scan_ir(stream_out,(unsigned short)(RD_IR_SZ*nb_chip),stream_in);		/* send the instruction stream */

    circulate_dr((unsigned short)(nb_chip*RD_BYPASS_SZ+instr_siz-1),stream_in); /* send the data stream */


    return get_stream(stream_in,(unsigned short)(read_i*RD_BYPASS_SZ),instr_siz);   /* return the old data (returned by the chip) */
}


/***************************************************************************************

	get_status_readout(map)

	chip  		map[number of chip]

	return : unsigned short status

	Description : 	this function gets the status register of the FPGA readout 
			and returns it (map describes the chain).


******************************************************************************************/
unsigned short get_status_readout(chip *map)
{
   return (unsigned short)get_one_readout(RD_STATUS,RD_STATUS_SZ,0,map);
}

/***************************************************************************************

	get_temp1_readout(map)

	chip  		map[number of chip]

	return : unsigned short status

	Description : 	this function gets the temperature register of the FPGA readout 
			and returns it (map describes the chain).


******************************************************************************************/
unsigned long get_temp1_readout(chip *map)
{
   return (unsigned long)get_one_readout(RD_TEMP1,RD_TEMP1_SZ,0,map);
}

/***************************************************************************************

	set_config_readout(data,map)

	unsigned short  data
	chip  		map[number of chip]

	return : unsigned short config

	Description : 	this function set the config register of the FPGA readout 
			and returns the last value (map describes the chain).


******************************************************************************************/
unsigned int set_config_readout(unsigned int data1,unsigned int data2,chip *map)
{
    unsigned short stream_in[MAX_LENGTH_STREAM],stream_out[MAX_LENGTH_STREAM];
    unsigned short nb_chip,i,iread,read_i=0;


    if((map_readout_nb(map))==0) return 0;
    nb_chip = map_nb(map);
    for (i=0,iread=0;i<nb_chip;i++)            /* find the readout in the chain */
        if (map[i]==READOUT)
	{
	    if (iread==0)
	    {
	        read_i=i;
                break;
	    }
	    else iread++;
	}

    init_stream(stream_out,(unsigned short)(RD_IR_SZ*nb_chip));	/* fill the output stream with BYPASS instructions */

    set_stream(stream_out,(unsigned short)(read_i*RD_IR_SZ),RD_CONFIG,RD_IR_SZ); /* put the instruction for the connexion in the output stream. */

    scan_ir(stream_out,(unsigned short)(RD_IR_SZ*nb_chip),stream_in);		/* send the instruction stream */

    set_stream(stream_out,(unsigned short)(read_i*RD_BYPASS_SZ),data1,32);	/* create the data stream to be sent */
    set_stream(stream_out,(unsigned short)(read_i*RD_BYPASS_SZ+32),data2,RD_CONFIG_SZ-32);	/* create the data stream to be sent */

    scan_dr(stream_out,(unsigned short)(nb_chip*RD_BYPASS_SZ+RD_CONFIG_SZ-1),stream_in); /* send the data stream */


    return get_stream(stream_in,(unsigned short)(read_i*RD_BYPASS_SZ),32);   /* return the old data (returned by the chip) */
}

/***************************************************************************************

	get_latchup_readout(map)

	chip  		map[number of chip]

	return : unsigned char latchup

	Description : 	this function gets the latchup status register of the FPGA readout 
			and returns it (map describes the chain).


******************************************************************************************/
unsigned char get_latchup_readout(chip *map)
{
   return (unsigned char)get_one_readout(RD_LATCHUP,RD_LATCHUP_SZ,0,map);
}

unsigned short get_nb_event_readout(chip *map)
{
   return (unsigned short)get_one_readout(RD_NB_EVENT,16,0,map);
}

void get_nb_event2_readout(unsigned short *data,chip *map)
{
    unsigned short stream_in[MAX_LENGTH_STREAM],stream_out[MAX_LENGTH_STREAM];
    unsigned short nb_chip,i;


    if((map_readout_nb(map))==0) return;
    nb_chip = map_nb(map);
   
    init_stream(stream_out,(unsigned short)(RD_IR_SZ*nb_chip));	/* fill the output stream with BYPASS instructions */

    set_stream(stream_out,(unsigned short)((nb_chip-1)*RD_IR_SZ),RD_NB_EVENT,RD_IR_SZ); /* put the instruction for the connexion in the output stream. */

    scan_ir(stream_out,(unsigned short)(RD_IR_SZ*nb_chip),stream_in);		/* send the instruction stream */

    circulate_dr((unsigned short)(nb_chip*RD_BYPASS_SZ+RD_NB_EVENT_SZ-1),stream_in); /* send the data stream */

    for(i=0;i<3;i++)
       data[i]=(unsigned short)get_stream(stream_in,(nb_chip-1)+i*(RD_NB_EVENT_SZ/3),(RD_NB_EVENT_SZ/3));
   
   return;
}
/***************************************************************************************

	set_chain_readout(data,map)

	unsigned char   data
	chip  		map[number of chip]

	return : unsigned char chain

	Description : 	this function sets the chain register of the FPGA readout
			and returns the last value. (map describes the chain).
			
			bit 0 : Trg_status_back (1 Slow control, 0 Daq)
			bit 1 : Fe_tdo_sel (1 whole chain, 0 Readout Board alone)


******************************************************************************************/
unsigned char set_chain_readout(unsigned char data,chip *map)
{
   return (unsigned char)send_one_readout(RD_CHAIN,RD_CHAIN_SZ,data,0,map);
}

/***************************************************************************************

	get_version_readout(version,map)

	unsigned short  version[3]
	chip  		map[number of chip]

	return : nothing

	Description : 	this function gets the version of the FPGA program. 
			(map describes the chain).


******************************************************************************************/
void get_version_readout(unsigned short *data,chip *map)
{
    unsigned short stream_in[MAX_LENGTH_STREAM],stream_out[MAX_LENGTH_STREAM];
    unsigned short nb_chip,i;


    if((map_readout_nb(map))==0) return;
    nb_chip = map_nb(map);
   
    init_stream(stream_out,(unsigned short)(RD_IR_SZ*nb_chip));	/* fill the output stream with BYPASS instructions */

    set_stream(stream_out,(unsigned short)((nb_chip-1)*RD_IR_SZ),RD_VERSION,RD_IR_SZ); /* put the instruction for the connexion in the output stream. */

    scan_ir(stream_out,(unsigned short)(RD_IR_SZ*nb_chip),stream_in);		/* send the instruction stream */

    circulate_dr((unsigned short)(nb_chip*RD_BYPASS_SZ+RD_VERSION_SZ-1),stream_in); /* send the data stream */

    for(i=0;i<3;i++)
       data[i]=(unsigned short)get_stream(stream_in,(nb_chip-1)+i*(RD_VERSION_SZ/3),(RD_VERSION_SZ/3));
   
   return;
}

/***************************************************************************************

	get_optic_var_readout(version,map)

	unsigned short  optic_var[3]
	chip  		map[number of chip]

	return : nothing

	Description : 	this function gets the value of the optic_var register of the FPGA program. 
			(map describes the chain).


******************************************************************************************/
void get_optic_var_readout(unsigned short *data,chip *map)
{
    unsigned short stream_in[MAX_LENGTH_STREAM],stream_out[MAX_LENGTH_STREAM];
    unsigned short nb_chip,i;


    if((map_readout_nb(map))==0) return;
    nb_chip = map_nb(map);
   
    init_stream(stream_out,(unsigned short)(RD_IR_SZ*nb_chip));	/* fill the output stream with BYPASS instructions */

    set_stream(stream_out,(unsigned short)((nb_chip-1)*RD_IR_SZ),RD_OPTIC_VAR,RD_IR_SZ); /* put the instruction for the connexion in the output stream. */

    scan_ir(stream_out,(unsigned short)(RD_IR_SZ*nb_chip),stream_in);		/* send the instruction stream */

    circulate_dr((unsigned short)(nb_chip*RD_BYPASS_SZ+RD_OPTIC_VAR_SZ-1),stream_in); /* send the data stream */

    for(i=0;i<3;i++)
       data[i]=(unsigned short)get_stream(stream_in,(nb_chip-1)+i*(RD_OPTIC_VAR_SZ/3),(RD_OPTIC_VAR_SZ/3));
   
   return;
}

/***************************************************************************************

	get_optic_fab_readout(version,map)

	unsigned short  optic_fab[4]
	chip  		map[number of chip]

	return : nothing

	Description : 	this function gets the value of the optic_var register of the FPGA program. 
			(map describes the chain).


******************************************************************************************/
void get_optic_fab_readout(unsigned short *data,chip *map)
{
    unsigned short stream_in[MAX_LENGTH_STREAM],stream_out[MAX_LENGTH_STREAM];
    unsigned short nb_chip,i;


    if((map_readout_nb(map))==0) return;
    nb_chip = map_nb(map);
   
    init_stream(stream_out,(unsigned short)(RD_IR_SZ*nb_chip));	/* fill the output stream with BYPASS instructions */

    set_stream(stream_out,(unsigned short)((nb_chip-1)*RD_IR_SZ),RD_OPTIC_FAB,RD_IR_SZ); /* put the instruction for the connexion in the output stream. */

    scan_ir(stream_out,(unsigned short)(RD_IR_SZ*nb_chip),stream_in);		/* send the instruction stream */

    circulate_dr((unsigned short)(nb_chip*RD_BYPASS_SZ+RD_OPTIC_FAB_SZ-1),stream_in); /* send the data stream */

    for(i=0;i<4;i++)
       data[i]=(unsigned short)get_stream(stream_in,(nb_chip-1)+i*(RD_OPTIC_FAB_SZ/4),(RD_OPTIC_FAB_SZ/4));
   
   return;
}
/***************************************************************************************

    get_daq_status_readout(status,map)

	unsigned short  status[3]
	chip  		map[number of chip]

	return : nothing

	Description : 	this function gets the value of the daq_status register of the FPGA program. 
			(map describes the chain).


******************************************************************************************/
void get_daq_status_readout(unsigned short *data,chip *map)
{
    unsigned short stream_in[MAX_LENGTH_STREAM],stream_out[MAX_LENGTH_STREAM];
    unsigned short nb_chip,i;


    if((map_readout_nb(map))==0) return;
    nb_chip = map_nb(map);
   
    init_stream(stream_out,(unsigned short)(RD_IR_SZ*nb_chip));	/* fill the output stream with BYPASS instructions */

    set_stream(stream_out,(unsigned short)((nb_chip-1)*RD_IR_SZ),RD_DAQ_STATUS,RD_IR_SZ); /* put the instruction for the connexion in the output stream. */

    scan_ir(stream_out,(unsigned short)(RD_IR_SZ*nb_chip),stream_in);		/* send the instruction stream */

    circulate_dr((unsigned short)(nb_chip*RD_BYPASS_SZ+RD_DAQ_STATUS_SZ-1),stream_in); /* send the data stream */

    for(i=0;i<3;i++)
       data[i]=(unsigned short)get_stream(stream_in,(nb_chip-1)+i*16,16);
       
    data[2]&=0xff;
   
   return;
}


/***************************************************************************************

	get_trigger_status_readout(nb)

	unsigned char  readout board number

	return : unsigned char status

	Description : 	this function gets the trigger status of the readout
			board in order to know if it is busy or not.


******************************************************************************************/
unsigned char get_trigger_status_readout(unsigned char nb)
{
   unsigned short status;

   if (nb>3) return 0;
   reada16(PIO1_IN,&status);
   return (status&(1<<nb))!=0;
}

unsigned char get_id_readout(chip *map)
{
   return (unsigned char)get_one_readout(RD_IDENT,RD_IDENT_SZ,0,map);
}

unsigned long set_pedestal_readout(unsigned long data,chip *map)
{
   return send_one_readout(RD_PEDESTAL,RD_PEDESTAL_SZ,data,0,map);
}

unsigned char set_adjust_readout(unsigned char data,chip *map)
{
   return send_one_readout(RD_ADJUST,RD_ADJUST_SZ,data,0,map);
}

