/* jtagA128C.c		Jtag A128C functions 
 *
 *	Institut de Recherche Subatomique de Strasbourg
 *
 *	Author :	David Bonnet
 *	Date :		22/03/00
 *
 */
 

#include "stdio.h"
#include "jtagCommon.h"
#include "jtagA128C.h"


/***************************************************************************************


	map_a128c_nb(map)
	
	chip 		map[number of chips]
	
	return : 	unsigned short a128c_nb
	
	Description :  	this function determines how many alice128c are in the chain, 
			according to the "map".


****************************************************************************************/

unsigned short map_a128c_nb(chip *map)
{
    unsigned short i=0,ial=0;
    
    for (;(map[i]!=END)&&(i<MAX_CHIPS);i++)
        if (map[i]==ALICE128C) ial++;
    if(i==MAX_CHIPS) ial=0;
    return ial;
}
  
/***************************************************************************************

	send_all_a128c(instruction,instr_siz,list_in,list_out,map)
	
	unsigned long 	instruction
	unsigned short 	instr_siz
	unsigned short 	*list_in
	unsigned short 	*list_out
	chip 		map[number of chips]
	
	return : 	nothing
	
	Description : 	this function send the instruction to all the a128c in the chain 
			(map descibes the chain), then send the data from list_out. the data returned
			by the a128cs are in list_in.
				

******************************************************************************************/

 
void send_all_a128c(unsigned long instruction,unsigned short instr_siz,
		     unsigned short *list_in,unsigned short *list_out,chip *map)
{
    unsigned short stream_in[MAX_LENGTH_STREAM];
    unsigned short stream_out[MAX_LENGTH_STREAM];
    unsigned short nb_chip,i,istr,idat;    

    if (map_a128c_nb(map)==0) return;
    nb_chip = map_nb(map);
#if OUT_PUT
    printf("\n%s, %d: %s start\n", __FILE__, __LINE__, __func__);
    print_map(map);
#endif      
    init_stream(stream_out,AL_IR_SZ*nb_chip);	/* fill the output stream with BYPASS instructions */
    
    for (i=0;i<nb_chip;i++)					/* put the instruction for the a128cs in */
        if (map[i]==ALICE128C)					/* the input stream. 			  */
	    set_stream(stream_out,i*AL_IR_SZ,instruction,AL_IR_SZ); 
	       
    scan_ir(stream_out,AL_IR_SZ*nb_chip,stream_in);		/* send the instruction stream */
    
    for (i=0,istr=0,idat=0;i<nb_chip;i++)			/* create the data stream to be sent */
        if (map[i]==ALICE128C)
	   {
	   set_stream(stream_out,istr,get_stream(list_out,idat,instr_siz),instr_siz);
	   idat+=instr_siz;
	   istr+=instr_siz;
	   }
	else istr+=AL_BYPASS_SZ; 
    
    scan_dr(stream_out,istr,stream_in);				/* send the data stream */
    
    for (i=0,istr=0,idat=0;i<nb_chip;i++)			/* put the data from stream_in into list_in */
        if (map[i]==ALICE128C)
	   {
	   set_stream(list_in,idat,get_stream(stream_in,istr,instr_siz),instr_siz);
	   idat+=instr_siz;
	   istr+=instr_siz;
	   }
	else istr+=AL_BYPASS_SZ;
    /*print_stream(list_in,map_a128c_nb(map)*instr_siz);*/
#if OUT_PUT
    printf("===========================================\n");
#endif 
    return;
}    
    
  
/***************************************************************************************

	send_one_a128c(instruction,instr_siz,data,a128c_nb,map)
	
	unsigned long 	instruction
	unsigned short 	instr_siz
	unsigned long 	data
	unsigned short 	a128c_nb
	chip 		map[number of chips]
	
	return : 	unsigned long old_data
	
	
	Description : 	this function send the instruction to the alice128c "a128c_nb" 
			(map descibes the chain), then send the data. the data returned
			by the alice128c (the old data) is returned by this function.
				

******************************************************************************************/


unsigned long send_one_a128c(unsigned long instruction,unsigned short instr_siz,
			      unsigned long data,unsigned short a128c_nb,chip *map)
{
    unsigned short stream_in[MAX_LENGTH_STREAM],stream_out[MAX_LENGTH_STREAM];
    unsigned short nb_chip,i,ial,al_i=0;    


    if (map_a128c_nb(map)==0) return 0;
    nb_chip = map_nb(map);
    
    for (i=0,ial=0;i<nb_chip;i++)            /* find the alice128C in the chain */ 
        if (map[i]==ALICE128C)
	{
	    if (ial==a128c_nb)
	    {
	        al_i=i; 
                break;
	    }
	    else ial++; 
	}
#if OUT_PUT
    printf("\n%s, %d: %s start\n", __FILE__, __LINE__, __func__);
    print_map(map);
#endif  	
    init_stream(stream_out,AL_IR_SZ*nb_chip);	/* fill the output stream with BYPASS instructions */
    
    set_stream(stream_out,al_i*AL_IR_SZ,instruction,AL_IR_SZ); /* put the instruction for the alice128C in the output stream. */
	    
    scan_ir(stream_out,AL_IR_SZ*nb_chip,stream_in);		/* send the instruction stream */
    
    set_stream(stream_out,al_i*AL_BYPASS_SZ,data,instr_siz);	/* create the data stream to be sent */
        
    scan_dr(stream_out,nb_chip*AL_BYPASS_SZ+instr_siz-1,stream_in); /* send the data stream */
    
    /*print_stream(stream_in,nb_chip*AL_BYPASS_SZ+instr_siz-1);*/
#if OUT_PUT
    printf("===========================================\n");
#endif       
    return get_stream(stream_in,al_i*AL_BYPASS_SZ,instr_siz);   /* return the old data (returned by the chip) */
}       

  
/***************************************************************************************

	get_all_a128c(instruction,instr_siz,list,map)
	
	unsigned long 	instruction
	unsigned short 	instr_siz
	unsigned short 	*list
	chip 		map[number of chips]
	
	return :	nothing
	
	
	Description : 	this function send the instruction to all the alice128c in the chain 
			(map descibes the chain), then get the data from the alice128Cs. 
				

******************************************************************************************/

 
void get_all_a128c(unsigned long instruction,unsigned short instr_siz,
		     unsigned short *list,chip *map)
{
    unsigned short stream[MAX_LENGTH_STREAM];
    unsigned short stream_in[MAX_LENGTH_STREAM];
    unsigned short nb_chip,i,istr,idat;    


    if (map_a128c_nb(map)==0) return;
    nb_chip = map_nb(map);
#if OUT_PUT
    printf("\n%s, %d: %s start\n", __FILE__, __LINE__, __func__);
    print_map(map);
#endif      
    init_stream(stream,AL_IR_SZ*nb_chip);	/* fill the output stream with BYPASS instructions */
    
    for (i=0;i<nb_chip;i++)					/* put the instruction for the alice128Cs in */
        if (map[i]==ALICE128C)					/* the input stream. 			  */
	    set_stream(stream,i*AL_IR_SZ,instruction,AL_IR_SZ); 
	       
    scan_ir(stream,AL_IR_SZ*nb_chip,stream_in);		/* send the instruction stream */
    
    circulate_dr(map_a128c_nb(map)*(instr_siz-1)+nb_chip,stream);	/* get the data stream */
    
    for (i=0,istr=0,idat=0;i<nb_chip;i++)			/* put the data from stream into list */
        if (map[i]==ALICE128C)
	   {
	   set_stream(list,idat,get_stream(stream,istr,instr_siz),instr_siz);
	   idat+=instr_siz;
	   istr+=instr_siz;
	   }
	else istr+=AL_BYPASS_SZ;
    /*print_stream(list,map_a128c_nb(map)*instr_siz);*/
#if OUT_PUT
    printf("===========================================\n");
#endif   
    return;
}    
    
  
/***************************************************************************************

	get_one_a128c(instruction,instr_siz,a128c_nb,map)
	
	unsigned long 	instruction
	unsigned short 	instr_siz
	unsigned short 	a128c_nb
	chip 		map[number of chips]
	
	return :	unsigned long data
	
	Description : 	this function send the instruction to the alice128C "a128c_nb" 
			(map descibes the chain), then get the data from this alice128C.
				

******************************************************************************************/


unsigned long get_one_a128c(unsigned long instruction,unsigned short instr_siz,
			      unsigned short a128c_nb,chip *map)
{
    unsigned short stream_in[MAX_LENGTH_STREAM],stream[MAX_LENGTH_STREAM];
    unsigned short nb_chip,i,ial,al_i=0;    


    if (map_a128c_nb(map)==0) return 0;
    nb_chip = map_nb(map);
    
    for (i=0,ial=0;i<nb_chip;i++)            /* find the alice128C in the chain */ 
        if (map[i]==ALICE128C)
	{
	    if (ial==a128c_nb)
	    {
	        al_i=i; 
                break;
	    }
	    else ial++; 
	}
#if OUT_PUT
    printf("\n%s, %d: %s start\n", __FILE__, __LINE__, __func__);
    print_map(map);
#endif  	
    init_stream(stream,AL_IR_SZ*nb_chip);	/* fill the output stream with BYPASS instructions */
    
    set_stream(stream,al_i*AL_IR_SZ,instruction,AL_IR_SZ); /* put the instruction for the alice128C in the output stream. */
	    
    scan_ir(stream,AL_IR_SZ*nb_chip,stream_in);		/* send the instruction stream */
            
    circulate_dr(nb_chip*AL_BYPASS_SZ+instr_siz-1,stream); /* get the data stream */
    
    /*print_stream(stream,nb_chip*AL_BYPASS_SZ+instr_siz-1);*/
#if OUT_PUT
    printf("===========================================\n");
#endif       
    return get_stream(stream,al_i*AL_BYPASS_SZ,instr_siz);   /* return the data */
}       

  
/***************************************************************************************

	set_pulse_all_a128c(level,map)
	
	unsigned char 	level
	chip 		map[number of chips]
	
	return :	nothing
	
	Description : 	this function set the pulse level of all the alice128Cs 
			(map describes the chain).
				

******************************************************************************************/
void set_pulse_all_a128c(unsigned char level,chip *map)
{
    unsigned short list_in[MAX_LENGTH_STREAM],list_out[MAX_LENGTH_STREAM];
    unsigned short nb_al,i;
    
    if((nb_al = map_a128c_nb(map))==0) return;
    
    for(i=0;i<nb_al;i++)    		   /* the same level for all the alice128Cs */
        set_stream(list_out,i*AL_PULSE_SZ,level,AL_PULSE_SZ);
    
    send_all_a128c(AL_PULSE,AL_PULSE_SZ,list_in,list_out,map);
    
}

  
/***************************************************************************************

	set_pulse_one_a128c(level,al_nb,map)
	
	unsigned char 	level
	chip 		map[number of chips]
	
	return :	nothing
	
	Description : 	this function set the pulse level of the alice128Cs "al_nb"
			(map describes the chain).
				

******************************************************************************************/
unsigned char set_pulse_one_a128c(unsigned char level,unsigned short al_nb,chip *map)
{
    return (unsigned char) send_one_a128c(AL_PULSE,AL_PULSE_SZ,(unsigned long) level,al_nb,map);
    
}

  
/***************************************************************************************

	set_bias_all_a128c(bias,map)
	
	unsigned char 	bias[number of a128c][number of bias registers (7)]
	unsigned char	old_bias[number of a128c][number of bias registers (7)]
	chip 		map[number of chips]
	
	return :	nothing
	
	Description : 	this function set the bias of all the alice128Cs 
			(map describes the chain).
			
	Note : 		bias[][0] : Ipreamp
			bias[][1] : Vpreamp
			bias[][2] : Ishaper
			bias[][3] : Vshaper
			bias[][4] : Iinbuff
			bias[][5] : Iouputbuff
			bias[][6] : Ilvds
				

******************************************************************************************/
void set_bias_all_a128c(unsigned char bias[][7],unsigned char old_bias[][7],chip *map)
{
    unsigned short stream_in[MAX_LENGTH_STREAM],stream_out[MAX_LENGTH_STREAM];
    unsigned short nb_chip,i,ial,j,istr;
        
    if(map_a128c_nb(map)==0) return;
    nb_chip = map_nb(map);
#if OUT_PUT
    printf("\n%s, %d: %s start\n", __FILE__, __LINE__, __func__);
    print_map(map);
#endif      
    init_stream(stream_out,AL_IR_SZ*nb_chip);	/* fill the output stream with BYPASS instructions */
    
    for (i=0;i<nb_chip;i++)					/* put the instruction for the a128cs in */
        if (map[i]==ALICE128C)					/* the input stream. 			  */
	    set_stream(stream_out,i*AL_IR_SZ,AL_BIAS,AL_IR_SZ); 
	       
    scan_ir(stream_out,AL_IR_SZ*nb_chip,stream_in);		/* send the instruction stream */
    
    for (i=0,istr=0,ial=0;i<nb_chip;i++)			/* create the data stream to be sent */
        if (map[i]==ALICE128C)
	   {
	      for (j=0;j<7;j++)
	      {
	         set_stream(stream_out,istr,bias[ial][j],AL_BIAS_SZ/7);
	         istr+=AL_BIAS_SZ/7;
	      }
	      ial++;
	   }
	else istr+=AL_BYPASS_SZ; 
    
    scan_dr(stream_out,istr,stream_in);				/* send the data stream */
    
    for (i=0,istr=0,ial=0;i<nb_chip;i++)			/* put the data from stream_in into old_bias[][] */
        if (map[i]==ALICE128C)
	{
	   for (j=0;j<7;j++)
	   {
	      old_bias[ial][j] = (unsigned char) get_stream(stream_in,istr,AL_BIAS_SZ/7);
	      istr+=AL_BIAS_SZ/7;
	   }
	   ial++;
	}
	else istr+=AL_BYPASS_SZ;
#if OUT_PUT
    printf("===========================================\n");
#endif   	
    return;
}

  
/***************************************************************************************

	set_bias_one_a128c(bias,a128c_nb,map)
	
	unsigned char 	bias[number of bias registers (7)]
	unsigned char 	old_bias[number of bias registers (7)]
	unsigned short	a128c_nb
	chip 		map[number of chips]
	
	return :	nothing
	
	Description : 	this function set the bias of all the alice128Cs 
			(map describes the chain).
			
	Note : 		bias[0] : Ipreamp
			bias[1] : Vpreamp
			bias[2] : Ishaper
			bias[3] : Vshaper
			bias[4] : Iinbuff
			bias[5] : Iouputbuff
			bias[6] : Ilvds
				

******************************************************************************************/
void set_bias_one_a128c(unsigned char bias[7],unsigned char old_bias[7],unsigned short a128c_nb,chip *map)
{
    unsigned short stream_in[MAX_LENGTH_STREAM],stream_out[MAX_LENGTH_STREAM];
    unsigned short nb_chip,i,ial,al_i=0,j;    


    if(map_a128c_nb(map)==0) return;
    nb_chip = map_nb(map);
    
    for (i=0,ial=0;i<nb_chip;i++)            /* find the alice128C in the chain */ 
        if (map[i]==ALICE128C)
	{
	    if (ial==a128c_nb)
	    {
	        al_i=i; 
                break;
	    }
	    else ial++; 
	}
#if OUT_PUT
    printf("\n%s, %d: %s start\n", __FILE__, __LINE__, __func__);
    print_map(map);
#endif  	
    init_stream(stream_out,AL_IR_SZ*nb_chip);	/* fill the output stream with BYPASS instructions */
    
    set_stream(stream_out,al_i*AL_IR_SZ,AL_BIAS,AL_IR_SZ); /* put the instruction for the alice128C in the output stream. */
	    
    scan_ir(stream_out,AL_IR_SZ*nb_chip,stream_in);		/* send the instruction stream */
    
    for (j=0;j<7;j++)
       set_stream(stream_out,al_i*AL_BYPASS_SZ+j*(AL_BIAS_SZ/7),bias[j],AL_BIAS_SZ/7);	/* create the data stream to be sent */
        
    scan_dr(stream_out,nb_chip*AL_BYPASS_SZ+AL_BIAS_SZ-1,stream_in); /* send the data stream */
     
     
    for (j=0;j<7;j++) 
       old_bias[j] = (unsigned char) get_stream(stream_in,al_i*AL_BYPASS_SZ+j*(AL_BIAS_SZ/7),AL_BIAS_SZ/7);   
#if OUT_PUT
    printf("===========================================\n");
#endif       
    return;   
   
}

  
/***************************************************************************************

	set_onoff_all_a128c(onoff,map)
	
	unsigned char 	onoff
	chip 		map[number of chips]
	
	return :	nothing
	
	Description : 	this function set the onoffEnable register of all the alice128Cs 
			(map describes the chain).
	
	Note	    : onoff=1 : intermediate buffers and ouput buffer switched on.
		      onoff=0 : intermediate buffers and ouput buffer switched off.
				

******************************************************************************************/
void set_onoff_all_a128c(unsigned char *onoff,chip *map)
{
    unsigned short list_in[MAX_LENGTH_STREAM],list_out[MAX_LENGTH_STREAM];
    unsigned short nb_al,i;
    
    if((nb_al = map_a128c_nb(map))==0) return;
    
    for(i=0;i<nb_al;i++) {   		   
        set_stream(list_out,i*AL_ONOFF_SZ,onoff[i],AL_ONOFF_SZ);
	}
    
    send_all_a128c(AL_ONOFF,AL_ONOFF_SZ,list_in,list_out,map);
    
}

  
/***************************************************************************************

	set_onoff_one_a128c(onoff,a128c_nb,map)
	
	unsigned char 	onoff
	unsigned short	a128c_nb
	chip 		map[number of chips]
	
	return :	nothing
	
	Description : 	this function set the onoffEnable register of the alice128C 'a128c_nb'  
			(map describes the chain).
	
	Note	    : onoff=1 : intermediate buffers and ouput buffer switched on.
		      onoff=0 : intermediate buffers and ouput buffer switched off.
				

******************************************************************************************/
unsigned char set_onoff_one_a128c(unsigned char onoff,unsigned short a128c_nb,chip *map)
{
    return (unsigned char) send_one_a128c(AL_ONOFF,AL_ONOFF_SZ,(unsigned long) onoff,a128c_nb,map);
}

  
/***************************************************************************************

	set_token_all_a128c(token,map)
	
	unsigned char 	token
	chip 		map[number of chips]
	
	return :	nothing
	
	Description : 	this function set the tokenEnable register of all the alice128Cs 
			(map describes the chain).
	
	Note	    : token=1 : Token enabled.
		      token=0 : TokenIN bypassed to TokenOUT.
				

******************************************************************************************/
void set_token_all_a128c(unsigned char *token,chip *map)
{
    unsigned short list_in[MAX_LENGTH_STREAM],list_out[MAX_LENGTH_STREAM];
    unsigned short nb_al,i;
    
    if((nb_al = map_a128c_nb(map))==0) return;
    for(i=0;i<nb_al;i++)  {  		   
        set_stream(list_out,i*AL_TOKEN_SZ,token[i],AL_TOKEN_SZ);
	}
    send_all_a128c(AL_TOKEN,AL_TOKEN_SZ,list_in,list_out,map);
}

  
/***************************************************************************************

	set_token_one_a128c(token,a128c_nb,map)
	
	unsigned char 	token
	unsigned short  a128c_nb
	chip 		map[number of chips]
	
	return :	nothing
	
	Description : 	this function set the tokenEnable register of the alice128Cs 'a128c_nb' 
			(map describes the chain).
	
	Note	    : token=1 : Token enabled.
		      token=0 : TokenIN bypassed to TokenOUT.
				

******************************************************************************************/
unsigned char set_token_one_a128c(unsigned char token,unsigned short a128c_nb,chip *map)
{
    return (unsigned char) send_one_a128c(AL_TOKEN,AL_TOKEN_SZ,(unsigned long) token,a128c_nb,map);
}

 
/***************************************************************************************

	set_test_one_a128c(channel,old_channel,a128c_nb,map)
	
	unsigned short 	channel[8]       8*16 = 128 = size of the Test register.
	unsigned short 	old_channel[8]
	unsigned short  a128c_nb
	chip 		map[number of chips]
	
	return :	nothing
	
	Description : 	this function set the Test register of the alice128Cs 'a128c_nb' 
			(map describes the chain).
					

******************************************************************************************/
void  set_test_one_a128c(unsigned short *channel,unsigned short *old_channel,unsigned short a128c_nb,chip *map)
{
    unsigned short stream_in[MAX_LENGTH_STREAM],stream_out[MAX_LENGTH_STREAM];
    unsigned short nb_chip,i,ial,al_i=0,j;    


    if(map_a128c_nb(map)==0) return;
    nb_chip = map_nb(map);
    
    for (i=0,ial=0;i<nb_chip;i++)            /* find the alice128C in the chain */ 
        if (map[i]==ALICE128C)
	{
	    if (ial==a128c_nb)
	    {
	        al_i=i; 
                break;
	    }
	    else ial++; 
	}
#if OUT_PUT
    printf("\n%s, %d: %s start\n", __FILE__, __LINE__, __func__);
    print_map(map);
#endif  	
    init_stream(stream_out,AL_IR_SZ*nb_chip);	/* fill the output stream with BYPASS instructions */
    
    set_stream(stream_out,al_i*AL_IR_SZ,AL_TEST,AL_IR_SZ); /* put the instruction for the alice128C in the output stream. */
	    
    scan_ir(stream_out,AL_IR_SZ*nb_chip,stream_in);		/* send the instruction stream */
    
    for (j=0;j<8;j++)	
       set_stream(stream_out,al_i*AL_BYPASS_SZ+j*16,channel[j],16);	/* create the data stream to be sent */
        
    scan_dr(stream_out,nb_chip*AL_BYPASS_SZ+AL_TEST_SZ-1,stream_in); /* send the data stream */
     
     
    for (j=0;j<8;j++) 
       old_channel[j] = (unsigned short) get_stream(stream_in,al_i*AL_BYPASS_SZ+j*16,16);   
#if OUT_PUT
    printf("===========================================\n");
#endif       
    return;   
   
}

void transp_chan_a128c(unsigned char channel,unsigned short a128c_nb,chip *map)
{
    unsigned short old_test[17],test[17];
    unsigned short stream_in[MAX_LENGTH_STREAM],stream_out[MAX_LENGTH_STREAM];
    unsigned short nb_chip,i,ial,al_i=0,j;    
    
    if(map_a128c_nb(map)==0) return;
    for (i=0;i<8;test[i++]=0);
    
    /*set_token_all_a128c(0,map);*/		/* set all the chips token in bypass mode */
    
    test[channel>>4]=(1<<(channel&15));
    
    set_test_one_a128c(test,old_test,a128c_nb,map); /*set the test register according to the 
    							selected channel */
    
    /* now set the huge read register (258 bits) */
    
    for (i=0;i<17;test[i++]=0);
    
    test[(channel+1)>>4]=(1<<((channel+1)&15)); /* set the onoff register bit */
    test[(channel+129)>>4]=(1<<((channel+129)&15)); /*set the readout register bit */
    test[16]+=2;   				/*set the outbuff control bit */
    test[0]+=1;
    nb_chip = map_nb(map);
    
    for (i=0,ial=0;i<nb_chip;i++)            /* find the alice128C in the chain */ 
        if (map[i]==ALICE128C)
	{
	    if (ial==a128c_nb)
	    {
	        al_i=i; 
                break;
	    }
	    else ial++; 
	}
#if OUT_PUT
    printf("\n%s, %d: %s start\n", __FILE__, __LINE__, __func__);
    print_map(map);
#endif  	
    init_stream(stream_out,AL_IR_SZ*nb_chip);	/* fill the output stream with BYPASS instructions */
    
    set_stream(stream_out,al_i*AL_IR_SZ,AL_READ,AL_IR_SZ); /* put the instruction for the alice128C in the output stream. */
	    
    scan_ir(stream_out,AL_IR_SZ*nb_chip,stream_in);		/* send the instruction stream */
    
    for (j=0;j<17;j++)	
       set_stream(stream_out,al_i*AL_BYPASS_SZ+j*16,test[j],16);	/* create the data stream to be sent */
        
    scan_dr(stream_out,nb_chip*AL_BYPASS_SZ+AL_READ_SZ-1,stream_in); /* send the data stream */
#if OUT_PUT
    printf("===========================================\n");
#endif       
    return;        
}
